#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
#
# Licensed under LGPLv3, see file LICENSE in this source tree.
"""
This module will start the Flask server 
"""
from threading import Thread
import weakref
import logging
from flask import Flask, render_template, jsonify, request
from flask_bootstrap import Bootstrap
import json
from bokeh.embed import server_document

from .templates import create_sidebar, create_card, update_notifications
from flask.logging import default_handler


class FlaskServer(Thread):

    # Init thread running server
    def __init__(self, network, port=8111, ip="0.0.0.0", *, daemon=True):
        Thread.__init__(self, daemon=daemon)
        self.flask_app = Flask(__name__)
        Bootstrap(self.flask_app)
        self._network_ref = weakref.ref(network)
        self.port = port
        self.ip = ip
        self.notifications_log = []
        self.notifications_list = ""
        self.config_flask_app()
        self.exitFlag = False

    @property
    def network(self):
        return self._network_ref()

    def run(self):
        self.process()

    def process(self):
        while not self.exitFlag:
            self.task()

    def startServer(self):
        self.flask_app.run(port=self.port, host="0.0.0.0")
        self.flask_app.logger.removeHandler(default_handler)

    def config_flask_app(self):
        @self.flask_app.route("/trends", methods=["GET"])
        def bkapp_trends_page():
            if self.network.registered_devices:
                script = server_document("http://{}:5006/trends".format(self.ip))
            else:
                script = "<div>No devices registered yet...</div>"
            return render_template(
                "trends.html",
                sidebar=create_sidebar(trends_class='class="active"'),
                bokeh_script=script,
                template="Flask",
            )

        #        @self.flask_app.route('/devices', methods=['GET'])
        #        def bkapp_devices_page():
        #            script = server_document('http://localhost:5006/devices')
        #            return render_template("embed.html", script=script, template="Flask")

        @self.flask_app.route("/notes", methods=["GET"])
        def bkapp_notes_page():
            script = server_document("http://{}:5006/notes".format(self.ip))
            return render_template("embed.html", script=script, template="Flask")

        @self.flask_app.route("/", methods=["GET"])
        def home_page():
            # Stat number of devices
            cnod = create_card(
                icon="ti-server",
                title="Number of devices",
                data=self.network.number_of_devices,
                id_data="devices",
                foot_icon="ti-reload",
                foot_data="Refresh to update",
            )
            cnot = create_card(
                icon="ti-bar-chart",
                title="Number of trends",
                data=self.network.number_of_registered_trends,
                id_data="trends",
                foot_icon="ti-timer",
                foot_data="Add trends !",
            )

            cnmn = create_card(
                icon="ti-plug",
                title="{} MSTP Networks".format(
                    len(self.network.network_stats["mstp_networks"])
                ),
                data="# {}".format(self.network.network_stats["print_mstpnetworks"]),
                id_data="mstpnetworks",
                foot_icon="ti-timer",
                foot_data="Last update : {}".format(
                    self.network.network_stats["timestamp"]
                ),
                id_foot_data="lastwhoisupdate",
            )

            notif = update_notifications(self.notifications_log, None)
            return render_template(
                "dashboard.html",
                sidebar=create_sidebar(dash_class='class="active"'),
                card_number_of_devices=cnod,
                card_number_of_mstp_networks=cnmn,
                card_number_of_trends=cnot,
                notifications=notif,
                template="Flask",
            )

        @self.flask_app.route("/dash_devices", methods=["GET"])
        def dashboard_devices_page():
            script = server_document("http://{}:5006/devices".format(self.ip))
            return render_template(
                "device_table.html",
                sidebar=create_sidebar(devices_class='class="active"'),
                bokeh_script=script,
                template="Flask",
            )

        @self.flask_app.route("/_dash_live_data", methods=["GET"])
        def dash_live_data():
            devices = self.network.number_of_devices
            trends = self.network.number_of_registered_trends
            return jsonify(
                number_of_devices=devices, number_of_registered_trends=trends
            )

        @self.flask_app.route("/_whois", methods=["GET"])
        def whois():
            self.notifications_list = update_notifications(
                self.notifications_log, "Sent a WhoIs Request"
            )
            self.network.discover()
            return jsonify(done="done")

        @self.flask_app.route("/_dash_live_stats", methods=["GET"])
        def dash_live_stats():
            stats = self.network.network_stats
            return jsonify(stats=stats)

        @self.flask_app.route("/_network_pie_chart", methods=["GET"])
        def net_pie_chart():
            stats = self.network.number_of_devices_per_network()
            return jsonify(stats=stats)

        @self.flask_app.route("/log", methods=["POST", "GET"])
        def log_page():
            return json.dumps(request.form)

    def task(self):
        try:
            self.startServer()
        except Exception as err:
            self._log.warning("Flask server already running", err)
            self.exitFlag = True

    def stop(self):
        self._log.debug("Trying to stop Bokeh Server")
        # self.bokeh_server.stop()
        self.p.terminate()
        self.exitFlag = True

    def beforeStop(self):
        """
        Action done when closing thread
        """
        pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Reinitialize.py - creation of ReinitializeDeviceRequest

"""
from ...core.io.Read import find_reason
from ..io.IOExceptions import (
    SegmentationNotSupported,
    ReadPropertyException,
    ReadPropertyMultipleException,
    NoResponseFromController,
    ApplicationNotStarted,
)
from ...core.utils.notes import note_and_log

# --- standard Python modules ---
import datetime as dt

# --- 3rd party modules ---
from bacpypes.pdu import Address, GlobalBroadcast
from bacpypes.primitivedata import Date, Time, CharacterString
from bacpypes.basetypes import DateTime
from bacpypes.apdu import (
    ReinitializeDeviceRequest,
    ReinitializeDeviceRequestReinitializedStateOfDevice,
    SimpleAckPDU,
)
from bacpypes.iocb import IOCB
from bacpypes.core import deferred


@note_and_log
class Reinitialize:
    """
    Mixin to support Reinitialize from BAC0 to other devices
    """

    def reinitialize(self, address=None, password=None, state="coldstart"):
        """
        Will send reinitialize request
        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        if not address:
            raise ValueError("Provide address for request")

        # build a request
        request = ReinitializeDeviceRequest()
        request.reinitializedStateOfDevice = ReinitializeDeviceRequestReinitializedStateOfDevice.enumerations[
            state
        ]
        request.pduDestination = Address(address)
        request.password = CharacterString(password)

        self._log.debug("{:>12} {}".format("- request:", request))

        iocb = IOCB(request)  # make an IOCB

        # pass to the BACnet stack
        deferred(self.this_application.request_io, iocb)

        # Unconfirmed request...so wait until complete
        iocb.wait()  # Wait for BACnet response

        if iocb.ioResponse:  # successful response
            apdu = iocb.ioResponse

            if not isinstance(apdu, SimpleAckPDU):  # expect an ACK
                self._log.warning("Not an ack, see debug for more infos.")
                self._log.debug(
                    "Not an ack. | APDU : {} / {}".format((apdu, type(apdu)))
                )
                return

        if iocb.ioError:  # unsuccessful: error/reject/abort
            apdu = iocb.ioError
            reason = find_reason(apdu)
            raise NoResponseFromController("APDU Abort Reason : {}".format(reason))

        self._log.info("Reinitialize request sent to device : {}".format(address))



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Match.py - verify a point's status matches its commanded value.

Example:
    Is a fan commanded to 'On' actually 'running'?
"""

# --- standard Python modules ---
# --- 3rd party modules ---

# --- this application's modules ---
from .TaskManager import Task
from ..core.utils.notes import note_and_log

# ------------------------------------------------------------------------------


@note_and_log
class Match(Task):
    """
    Match two properties of a BACnet Object (i.e. a point status with its command).
    """

    def __init__(self, status=None, command=None, delay=5, name=None):
        self._log.debug(
            "Creating Match task for {} and {}. Delay : {}".format(
                command, status, delay
            )
        )
        if not name:
            name = "Match on " + status.properties.name
        self.command = command
        self.status = status
        Task.__init__(self, delay=delay, name=name)

    def task(self):
        try:
            if self.status.history[-1] != self.command.history[-1]:
                _val = (
                    self.command.history[-1].split(":")[1]
                    if ":" in self.command.history[-1]
                    else self.command.history[-1]
                )
                self.status._setitem(_val)
        except Exception:
            self._log.error(
                "Something wrong matching {} and {}... try again next time...".format(
                    self.command, self.status
                )
            )

    def stop(self):
        self.status._setitem("auto")
        super().stop()


@note_and_log
class Match_Value(Task):
    """
    Verify a point's Present_Value equals the given value after a delay of X seconds.
    Thus giving the BACnet controller (and connected equipment) time to respond to the
    command.
    
        Match_Value(On, <AI:1>, 5)
        
    i.e. Does Fan value = On after 5 seconds. 
    """

    def __init__(
        self, value=None, point=None, delay=5, name=None, use_last_value=False
    ):
        self._log.debug("Creating MatchValue task for {} and {}".format(value, point))
        self.value = value
        self.point = point
        self.use_last_value = use_last_value
        if not name:
            name = "Match_Value on " + point.properties.name
        Task.__init__(self, delay=delay, name=name)

    def task(self):
        try:
            if self.use_last_value:
                _point = self.point.lastValue
            else:
                _point = self.point.value
            value = self.value() if hasattr(self.value, "__call__") else self.value
            if value != _point:
                self.point._set(value)
        except Exception:
            self._log.error(
                "Something is wrong matching {} and {}... try again next time".format(
                    self.value, self.point
                )
            )

    def stop(self):
        self.point._set("auto")
        super().stop()



# -*- coding: utf-8 -*-
"""
Notes and logger decorator to be used on class
This will add a "notes" object to the class and will allow
logging feature at the same time.
Goal is to be able to access quickly to important informations for
the web interface.
"""
# --- standard Python modules ---
from collections import namedtuple
from datetime import datetime
import logging
from logging import FileHandler
import sys

import os
from os.path import expanduser, join

# --- 3rd party modules ---
try:
    import pandas as pd

    _PANDAS = True
except ImportError:
    _PANDAS = False


class LogList:
    LOGGERS = []


def convert_level(level):
    if not level:
        return None
    _valid_levels = [
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ]
    if level in _valid_levels:
        return level
    if level.lower() == "info":
        return logging.INFO
    elif level.lower() == "debug":
        return logging.DEBUG
    elif level.lower() == "warning":
        return logging.WARNING
    elif level.lower() == "error":
        return logging.ERROR
    elif level.lower() == "critical":
        return logging.CRITICAL
    raise ValueError(
        "Wrong log level use one of the following : {}".format(_valid_levels)
    )


def update_log_level(
    level=None, *, log_file=None, stderr=None, stdout=None, log_this=True
):
    """
    Typical usage ::
        # Silence (use CRITICAL so not much messages will be sent)
        BAC0.log_level('silence')
        # Verbose
        BAC0.log_level('info')
        # Default, Info on console....but Warning in file
        BAC0.log_level(file='warning', stdout='info', stderr='critical')
        # Debug in file and console... this is a bad idea as the console will be filled
        BAC0.log_level(file='debug', stdout='debug', stderr='critical')
           
        # Preferably, debug in the file, keep console limited to info
        BAC0.log_level('debug')
        # OR
        BAC0.log_level(file='debug', stdout='info', stderr='critical')   
        

    Giving only one parameter will set file and console to the same level.
    I tend to keep stderr CRITICAL
     
    """
    update_log_file_lvl = False
    update_stderr_lvl = False
    update_stdout_lvl = False

    if level:
        logging.getLogger("BAC0_Root.BAC0.scripts.Base.Base").disabled = False
        if level.lower() == "silence":
            log_file_lvl = logging.CRITICAL
            stderr_lvl = logging.CRITICAL
            stdout_lvl = logging.CRITICAL
            update_log_file_lvl = True
            update_stderr_lvl = True
            update_stdout_lvl = True
            logging.getLogger("BAC0_Root.BAC0.scripts.Base.Base").disabled = True
        elif level.lower() == "default":
            log_file_lvl = logging.WARNING
            stderr_lvl = logging.CRITICAL
            stdout_lvl = logging.INFO
            update_log_file_lvl = True
            update_stderr_lvl = True
            update_stdout_lvl = True
        elif level.lower() == "debug":
            log_file_lvl = logging.DEBUG
            stdout_lvl = logging.INFO
            update_log_file_lvl = True
            update_stdout_lvl = True
        else:
            level = convert_level(level)
            log_file_lvl = level
            stdout_lvl = level
            update_log_file_lvl = True
            update_stdout_lvl = True

    else:
        if log_file:
            log_file_lvl = convert_level(log_file)
            update_log_file_lvl = True
        if stderr:
            stderr_lvl = convert_level(stderr)
            update_stderr_lvl = True
        if stdout:
            stdout_lvl = convert_level(stdout)
            update_stdout_lvl = True

    # Choose Base as logger for this task
    if log_this:
        BAC0_logger = logging.getLogger("BAC0_Root.BAC0.scripts.Base.Base")

    for each in LogList.LOGGERS:
        for handler in each.handlers:
            if update_log_file_lvl and handler.get_name() == "file_handler":
                handler.setLevel(log_file_lvl)
                if log_this:
                    BAC0_logger.warning(
                        "Changed log level of file to {}".format(
                            logging.getLevelName(log_file_lvl)
                        )
                    )
            elif update_stdout_lvl and handler.get_name() == "stdout":
                handler.setLevel(stdout_lvl)
                if log_this:
                    BAC0_logger.warning(
                        "Changed log level of console stdout to {}".format(
                            logging.getLevelName(stdout_lvl)
                        )
                    )
            elif update_stderr_lvl and handler.get_name() == "stderr":
                handler.setLevel(stderr_lvl)
                if log_this:
                    BAC0_logger.warning(
                        "Changed log level of console stderr to {}".format(
                            logging.getLevelName(stderr_lvl)
                        )
                    )


def note_and_log(cls):
    """
    This will be used as a decorator on class to activate
    logging and store messages in the variable cls._notes
    This will allow quick access to events in the web app.
    A note can be added to cls._notes without logging if passing
    the argument log=false to function note()
    Something can be logged without addind a note using function log()
    """
    if hasattr(cls, "DEBUG_LEVEL"):
        if cls.DEBUG_LEVEL == "debug":
            file_level = logging.DEBUG
            console_level = logging.DEBUG
        elif cls.DEBUG_LEVEL == "info":
            file_level = logging.INFO
            console_level = logging.INFO
    else:
        file_level = logging.WARNING
        console_level = logging.INFO
    # Notes object
    cls._notes = namedtuple("_notes", ["timestamp", "notes"])
    cls._notes.timestamp = []
    cls._notes.notes = []

    # Defining log object
    cls.logname = "{} | {}".format(cls.__module__, cls.__name__)
    cls._log = logging.getLogger("BAC0_Root.{}.{}".format(cls.__module__, cls.__name__))

    # Set level to debug so filter is done by handler
    cls._log.setLevel(logging.DEBUG)

    # Console Handler
    ch = logging.StreamHandler(sys.stderr)
    ch.set_name("stderr")
    ch.setLevel(logging.CRITICAL)

    ch2 = logging.StreamHandler(sys.stdout)
    ch2.set_name("stdout")
    ch2.setLevel(console_level)

    formatter = logging.Formatter("{asctime} - {levelname:<8}| {message}", style="{")

    # File Handler
    _PERMISSION_TO_WRITE = True
    logUserPath = expanduser("~")
    logSaveFilePath = join(logUserPath, ".BAC0")

    logFile = join(logSaveFilePath, "BAC0.log")
    if not os.path.exists(logSaveFilePath):
        try:
            os.makedirs(logSaveFilePath)
        except:
            _PERMISSION_TO_WRITE = False
    if _PERMISSION_TO_WRITE:
        fh = FileHandler(logFile)
        fh.set_name("file_handler")
        fh.setLevel(file_level)
        fh.setFormatter(formatter)

    ch.setFormatter(formatter)
    ch2.setFormatter(formatter)
    # Add handlers the first time only...
    if not len(cls._log.handlers):
        if _PERMISSION_TO_WRITE:
            cls._log.addHandler(fh)
        cls._log.addHandler(ch)
        cls._log.addHandler(ch2)

    LogList.LOGGERS.append(cls._log)

    def log_title(self, title, args=None, width=35):
        cls._log.debug("")
        cls._log.debug("#" * width)
        cls._log.debug("# {}".format(title))
        cls._log.debug("#" * width)
        if args:
            cls._log.debug("{!r}".format(args))
            cls._log.debug("#" * 35)

    def log_subtitle(self, subtitle, args=None, width=35):
        cls._log.debug("")
        cls._log.debug("=" * width)
        cls._log.debug("{}".format(subtitle))
        cls._log.debug("=" * width)
        if args:
            cls._log.debug("{!r}".format(args))
            cls._log.debug("=" * width)

    def log(self, note, *, level=logging.DEBUG):
        """
        Add a log entry...no note
        """
        if not note:
            raise ValueError("Provide something to log")
        note = "{} | {}".format(cls.logname, note)
        cls._log.log(level, note)

    def note(self, note, *, level=logging.INFO, log=True):
        """
        Add note to the object. By default, the note will also
        be logged
        :param note: (str) The note itself
        :param level: (logging.level)
        :param log: (boolean) Enable or disable logging of note
        """
        if not note:
            raise ValueError("Provide something to log")
        note = "{} | {}".format(cls.logname, note)
        cls._notes.timestamp.append(datetime.now())
        cls._notes.notes.append(note)
        if log:
            cls.log(level, note)

    @property
    def notes(self):
        """
        Retrieve notes list as a Pandas Series
        """
        if not _PANDAS:
            return dict(zip(self._notes.timestamp, self._notes.notes))
        return pd.Series(self._notes.notes, index=self._notes.timestamp)

    def clear_notes(self):
        """
        Clear notes object
        """
        cls._notes.timestamp = []
        cls._notes.notes = []

    # Add the functions to the decorated class
    cls.clear_notes = clear_notes
    cls.note = note
    cls.notes = notes
    cls.log = log
    cls.log_title = log_title
    cls.log_subtitle = log_subtitle
    return cls



from bacpypes.object import (
    AnalogValueObject,
    CharacterStringValueObject,
    AnalogInputObject,
    AnalogOutputObject,
    BinaryInputObject,
    BinaryOutputObject,
    BinaryValueObject,
    DateValueObject,
    DateTimeValueObject,
    MultiStateInputObject,
    MultiStateOutputObject,
    MultiStateValueObject,
)
from bacpypes.basetypes import (
    EngineeringUnits,
    BinaryPV,
    Polarity,
    Boolean,
    EventState,
    Date,
    DateTime,
    Unsigned,
)
from bacpypes.constructeddata import ArrayOf
from bacpypes.primitivedata import CharacterString, Real
from .object import ObjectFactory

"""
Those models are opiniated versions of BACnet objects.
They are meant to supply an easy mechanism to build BACnet objects
that will be served by the local device of BAC0.

You can indeed use this "method" and modify it to fill your needs.
"""


def _create(definition, **kwargs):
    _definition = {
        "name": None,
        "objectType": None,
        "instance": None,
        "properties": {},
        "description": None,
        "presentValue": None,
        "is_commandable": False,
        "relinquish_default": None,
    }
    _definition.update(definition)
    for k, v in kwargs.items():
        if k == "properties":
            for _k, _v in v.items():
                _definition[k][_k] = _v
        _definition[k] = v
    return ObjectFactory.from_dict(_definition)


def make_state_text(list_of_string):
    _arr = ArrayOf(CharacterString)
    _lst = [CharacterString(each) for each in list_of_string]
    return _arr(_lst)


def set_default_if_not_provided(prop, default_prop, **kwargs):
    try:
        _prop = kwargs[prop]
    except KeyError:
        _prop = default_prop
    return _prop


def analog(**kwargs):
    definition = {
        "instance": 0,
        "description": "No description",
        "properties": {
            "units": "percent",
            "eventState": EventState(),
            "covIncrement": 0.15,
        },
        "presentValue": 0,
        "is_commandable": False,
        "relinquish_default": 0,
    }
    return _create(definition, **kwargs)


def analog_input(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "AI", **kwargs)
    kwargs["objectType"] = AnalogInputObject
    kwargs["instance"] = 0
    # kwargs['is_commandable'] = True # Futur

    return analog(**kwargs)


def analog_output(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "AO", **kwargs)
    kwargs["objectType"] = AnalogOutputObject
    kwargs["instance"] = 0
    kwargs["is_commandable"] = True

    return analog(**kwargs)


def analog_value(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "AV", **kwargs)
    kwargs["objectType"] = AnalogValueObject
    kwargs["instance"] = 0
    kwargs["is_commandable"] = True
    return analog(**kwargs)


def binary(**kwargs):
    definition = {
        "instance": 0,
        "description": "No description",
        "properties": {"outOfService": Boolean(False), "eventState": EventState()},
        "presentValue": "inactive",
        "is_commandable": False,
        "relinquish_default": "inactive",
    }
    return _create(definition, **kwargs)


def binary_input(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "BI", **kwargs)
    kwargs["objectType"] = BinaryInputObject
    try:
        kwargs["properties"].update({"polarity": Polarity()})
    except KeyError:
        kwargs["properties"] = {"polarity": Polarity()}
    # kwargs['is_commandable'] = True # Futur

    return binary(**kwargs)


def binary_output(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "BO", **kwargs)
    kwargs["objectType"] = BinaryOutputObject
    kwargs["is_commandable"] = True
    try:
        kwargs["properties"].update({"polarity": Polarity()})
    except KeyError:
        kwargs["properties"] = {"polarity": Polarity()}
    return binary(**kwargs)


def binary_value(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "BV", **kwargs)
    kwargs["objectType"] = BinaryValueObject
    kwargs["is_commandable"] = True

    return binary(**kwargs)


def multistate(**kwargs):
    default_states = make_state_text(["Off", "On"])
    definition = {
        "instance": 0,
        "properties": {
            "eventState": EventState(),
            "stateText": default_states,
            # "numberOfStates": Unsigned(2),
        },
        "description": "No description",
        "presentValue": Unsigned(1),
        "is_commandable": False,
        "relinquish_default": Unsigned(1),
    }
    return _create(definition, **kwargs)


def multistate_input(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "MSI", **kwargs)
    kwargs["objectType"] = MultiStateInputObject
    kwargs["is_commandable"] = True
    kwargs["relinquish_default"] = Unsigned(1)

    return multistate(**kwargs)


def multistate_output(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "MSO", **kwargs)
    kwargs["objectType"] = MultiStateOutputObject
    kwargs["is_commandable"] = True
    kwargs["relinquish_default"] = Unsigned(1)

    return multistate(**kwargs)


def multistate_value(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "MSV", **kwargs)
    kwargs["objectType"] = MultiStateValueObject
    kwargs["is_commandable"] = True
    kwargs["relinquish_default"] = Unsigned(1)

    return multistate(**kwargs)


def temperature_input(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "TEMP_INPUT", **kwargs)
    kwargs["properties"] = {"units": "degreesCelsius"}
    return analog_input(**kwargs)


def temperature_value(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "TEMP_VALUE", **kwargs)
    kwargs["properties"] = {"units": "degreesCelsius"}
    return analog_value(**kwargs)


def humidity_input(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "HUM_INPUT", **kwargs)
    kwargs["properties"] = {"units": "percentRelativeHumidity"}
    return analog_input(**kwargs)


def humidity_value(**kwargs):
    kwargs["name"] = set_default_if_not_provided("name", "HUM_VALUE", **kwargs)
    kwargs["properties"] = {"units": "percentRelativeHumidity"}
    return analog_value(**kwargs)


def character_string(**kwargs):
    definition = {
        "name": "CS_VALUE",
        "objectType": CharacterStringValueObject,
        "instance": 0,
        "description": "No description",
        "presentValue": "",
        "properties": {},
        "is_commandable": False,
        "relinquish_default": "inactive",
    }
    return _create(definition, **kwargs)


def date_value(**kwargs):
    definition = {
        "name": "DATE_VALUE",
        "objectType": DateValueObject,
        "instance": 0,
        "description": "No description",
        "presentValue": Date(),
        "properties": {},
        "is_commandable": False,
        "relinquish_default": "inactive",
    }
    return _create(definition, **kwargs)


def datetime_value(**kwargs):
    definition = {
        "name": "DATETIME_VALUE",
        "objectType": DateTimeValueObject,
        "instance": 0,
        "description": "No description",
        "presentValue": DateTime(),
        "properties": {},
        "is_commandable": False,
        "relinquish_default": "inactive",
    }
    return _create(definition, **kwargs)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
ReadWriteScript - extended version of BasicScript.py

As everything is handled by the BasicScript, select the additional features you want::

    # Create a class that implements a basic script with read and write functions
    from BAC0.scripts.BasicScript import BasicScript
    from BAC0.core.io.Read import ReadProperty
    from BAC0.core.io.Write import WriteProperty
    class ReadWriteScript(BasicScript,ReadProperty,WriteProperty)

Once the class is created, create the local object and use it::

    bacnet = ReadWriteScript(localIPAddr = '192.168.1.10')
    bacnet.read('2:5 analogInput 1 presentValue)

"""
# --- standard Python modules ---
import time
from datetime import datetime
import weakref
from collections import namedtuple


# --- this application's modules ---
from ..scripts.Base import Base

from ..core.io.Read import ReadProperty
from ..core.io.Write import WriteProperty
from ..core.functions.GetIPAddr import HostIP
from ..core.functions.Discover import Discover
from ..core.functions.TimeSync import TimeSync
from ..core.functions.Reinitialize import Reinitialize
from ..core.functions.DeviceCommunicationControl import DeviceCommunicationControl
from ..core.functions.cov import CoV
from ..core.functions.Schedule import Schedule
from ..core.io.Simulate import Simulation
from ..core.devices.Points import Point
from ..core.devices.Device import RPDeviceConnected, RPMDeviceConnected
from ..core.devices.Trends import TrendLog
from ..core.devices.Virtuals import VirtualPoint
from ..core.utils.notes import note_and_log
from ..core.io.IOExceptions import (
    NoResponseFromController,
    UnrecognizedService,
    Timeout,
)
from ..tasks.RecurringTask import RecurringTask
from ..tasks.UpdateCOV import Update_local_COV

from ..infos import __version__ as version

from bacpypes.pdu import Address

# ------------------------------------------------------------------------------


@note_and_log
class Lite(
    Base,
    Discover,
    ReadProperty,
    WriteProperty,
    Simulation,
    TimeSync,
    Reinitialize,
    DeviceCommunicationControl,
    CoV,
    Schedule,
):
    """
    Build a BACnet application to accept read and write requests.
    [Basic Whois/IAm functions are implemented in parent BasicScript class.]
    Once created, execute a whois() to build a list of available controllers.
    Initialization requires information on the local device.

    :param ip='127.0.0.1': Address must be in the same subnet as the BACnet network
        [BBMD and Foreign Device - not supported]

    """

    def __init__(
        self,
        ip=None,
        port=None,
        mask=None,
        bbmdAddress=None,
        bbmdTTL=0,
        bdtable=None,
        ping=True,
        **params
    ):
        self._log.info(
            "Starting BAC0 version {} ({})".format(
                version, self.__module__.split(".")[-1]
            )
        )
        self._log.info("Use BAC0.log_level to adjust verbosity of the app.")
        self._log.info("Ex. BAC0.log_level('silence') or BAC0.log_level('error')")

        self._log.debug("Configurating app")
        self._registered_devices = weakref.WeakValueDictionary()

        # Ping task will deal with all registered device and disconnect them if they do not respond.

        self._ping_task = RecurringTask(
            self.ping_registered_devices, delay=10, name="Ping Task"
        )
        if ping:
            self._ping_task.start()

        if ip is None:
            host = HostIP(port)
            ip_addr = host.address
        else:
            try:
                ip, subnet_mask_and_port = ip.split("/")
                try:
                    mask, port = subnet_mask_and_port.split(":")
                except ValueError:
                    mask = subnet_mask_and_port
            except ValueError:
                ip = ip

            if not mask:
                mask = 24
            if not port:
                port = 47808
            ip_addr = Address("{}/{}:{}".format(ip, mask, port))
        self._log.info("Using ip : {ip_addr}".format(ip_addr=ip_addr))
        Base.__init__(
            self,
            localIPAddr=ip_addr,
            bbmdAddress=bbmdAddress,
            bbmdTTL=bbmdTTL,
            bdtable=bdtable,
            **params
        )

        self.bokehserver = False
        self._points_to_trend = weakref.WeakValueDictionary()

        # Announce yourself
        self.iam()

        # Do what's needed to support COV
        self._update_local_cov_task = namedtuple(
            "_update_local_cov_task", ["task", "running"]
        )
        self._update_local_cov_task.task = Update_local_COV(
            self, delay=1, name="Update Local COV Task"
        )
        self._update_local_cov_task.task.start()
        self._update_local_cov_task.running = True
        self._log.info("Update Local COV Task started")

    @property
    def known_network_numbers(self):
        """
        This function will read the table of known network numbers gathered by the 
        NetworkServiceElement. It will also look into the discoveredDevices property
        and add any network number that would not be in the NSE table.
        """
        if self.discoveredDevices:
            for each in self.discoveredDevices:
                addr, instance = each
                net = addr.split(":")[0] if ":" in addr else None
                if net:
                    self.this_application.nse._learnedNetworks.add(int(net))

        return self.this_application.nse._learnedNetworks

    def discover(
        self, networks="known", limits=(0, 4194303), global_broadcast=False, reset=False
    ):
        """
        Discover is meant to be the function used to explore the network when we
        connect.
        It will trigger whois request using the different options provided with 
        parameters.

        By default, a local broadcast will be used. This is required as in big 
        BACnet network, global broadcast can lead to network flood and loss of data.

        If not parameters are given, BAC0 will try to :

            * Find the network on which it is
            * Find routers for other networks (accessible via local broadcast)
            * Detect "known networks"
            * Use the list of known networks and create whois request to find all devices on those networks

        This should be sufficient for most cases.

        Once discovery is done, user may access the list of "discovered devices" using ::

            bacnet.discoveredDevices

        :param networks (list, integer) : A simple integer or a list of integer
            representing the network numbers used to issue whois request.

        :param limits (tuple) : tuple made of 2 integer, the low limit and the high
            limit. Those are the device instances used in the creation of the
            whois request. Min : 0 ; Max : 4194303

        :param global_broadcast (boolean) : If set to true, a global braodcast
            will be used for the whois. Use with care.
        """
        if reset:
            self.discoveredDevices = None
        found = []
        _networks = []
        deviceInstanceRangeLowLimit, deviceInstanceRangeHighLimit = limits
        # Try to find on which network we are
        self.what_is_network_number()
        # Try to find local routers...
        self.whois_router_to_network()
        self._log.info("Found those networks : {}".format(self.known_network_numbers))

        if networks:
            if isinstance(networks, list):
                # we'll make multiple whois...
                for network in networks:
                    if network < 65535:
                        _networks.append(network)
            elif networks == "known":
                _networks = self.known_network_numbers
            else:
                if networks < 65535:
                    _networks.append(networks)

        if _networks:
            for network in _networks:
                self._log.info("Discovering network {}".format(network))
                _res = self.whois(
                    "{}:* {} {}".format(
                        network,
                        deviceInstanceRangeLowLimit,
                        deviceInstanceRangeHighLimit,
                    ),
                    global_broadcast=global_broadcast,
                )
                for each in _res:
                    found.append(each)

        else:
            self._log.info(
                "No BACnet network found, attempting a simple whois using provided device instances limits ({} - {})".format(
                    deviceInstanceRangeLowLimit, deviceInstanceRangeHighLimit
                )
            )
            _res = self.whois(
                "{} {}".format(
                    deviceInstanceRangeLowLimit, deviceInstanceRangeHighLimit
                ),
                global_broadcast=global_broadcast,
            )
            for each in _res:
                found.append(each)
        return found

    def register_device(self, device):
        oid = id(device)
        self._registered_devices[oid] = device

    def ping_registered_devices(self):
        """
        Registered device on a network (self) are kept in a list (registered_devices). 
        This function will allow pinging thoses device regularly to monitor them. In case
        of disconnected devices, we will disconnect the device (which will save it). Then 
        we'll ping again until reconnection, where the device will be bring back online.

        To permanently disconnect a device, an explicit device.disconnect(unregister=True [default value])
        will be needed. This way, the device won't be in the registered_devices list and 
        BAC0 won't try to ping it.
        """
        for each in self.registered_devices:
            if isinstance(each, RPDeviceConnected) or isinstance(
                each, RPMDeviceConnected
            ):
                device_connected = True
            else:
                device_connected = False

            respond_to_ping = True
            try:
                self._log.debug(
                    "Ping {}|{}".format(each.properties.name, each.properties.address)
                )
                device_id = each.properties.device_id
                addr = each.properties.address
                self.read("{} device {} objectName".format(addr, device_id))
            except NoResponseFromController:
                respond_to_ping = False

            if not respond_to_ping and device_connected:
                self._log.warning(
                    "{}|{} is offline, disconnecting it.".format(
                        each.properties.name, each.properties.address
                    )
                )
                each.disconnect(unregister=False)
            elif not device_connected and respond_to_ping:
                self._log.info(
                    "{}|{} is back online, reconnecting.".format(
                        each.properties.name, each.properties.address
                    )
                )
                each.connect(network=self)

    @property
    def registered_devices(self):
        """
        Devices that have been created using BAC0.device(args)
        """
        return list(self._registered_devices.values())

    def unregister_device(self, device):
        """
        Remove from the registered list
        """
        oid = id(device)
        try:
            del self._registered_devices[oid]
        except KeyError:
            pass

    def add_trend(self, point_to_trend):
        """
        Add point to the list of histories that will be handled by Bokeh

        Argument provided must be of type Point or TrendLog
        ex. bacnet.add_trend(controller['point_name'])
        """
        if (
            isinstance(point_to_trend, Point)
            or isinstance(point_to_trend, TrendLog)
            or (isinstance(point_to_trend, VirtualPoint))
        ):
            oid = id(point_to_trend)
            self._points_to_trend[oid] = point_to_trend
        else:
            raise TypeError("Please provide point containing history")

    def remove_trend(self, point_to_remove):
        """
        Remove point from the list of histories that will be handled by Bokeh

        Argument provided must be of type Point or TrendLog
        ex. bacnet.remove_trend(controller['point_name'])
        """
        if (
            isinstance(point_to_remove, Point)
            or isinstance(point_to_remove, TrendLog)
            or isinstance(point_to_remove, VirtualPoint)
        ):
            oid = id(point_to_remove)
        else:
            raise TypeError("Please provide point or trendLog containing history")
        if oid in self._points_to_trend.keys():
            del self._points_to_trend[oid]

    @property
    def devices(self):
        """
        This property will create a good looking table of all the discovered devices
        seen on the network. 

        For that, some requests will be sent over the network to look for name, 
        manufacturer, etc and in big network, this could be a long process.
        """
        lst = []
        for device in list(self.discoveredDevices):
            try:
                deviceName, vendorName = self.readMultiple(
                    "{} device {} objectName vendorName".format(device[0], device[1])
                )
            except (UnrecognizedService, ValueError):
                self._log.warning(
                    "Unrecognized service for {} | {}".format(device[0], device[1])
                )
                try:
                    deviceName = self.read(
                        "{} device {} objectName".format(device[0], device[1])
                    )
                    vendorName = self.read(
                        "{} device {} vendorName".format(device[0], device[1])
                    )
                except NoResponseFromController:
                    self._log.warning("No response from {}".format(device))
                    continue
            except (NoResponseFromController, Timeout):
                self._log.warning("No response from {}".format(device))
                continue
            lst.append((deviceName, vendorName, device[0], device[1]))
        return lst

    @property
    def trends(self):
        """
        This will present a list of all registered trends used by Bokeh Server
        """
        return list(self._points_to_trend.values())

    def disconnect(self):
        self._log.debug("Disconnecting")
        for each in self.registered_devices:
            each.disconnect()
        super().disconnect()

    def __repr__(self):
        return "Bacnet Network using ip {} with device id {}".format(
            self.localIPAddr, self.Boid
        )

    def __getitem__(self, boid_or_localobject):
        item = self.this_application.objectName[boid_or_localobject]
        if item is None:
            for device in self._registered_devices:
                if str(device.properties.device_id) == str(boid_or_localobject):
                    return device
            self._log.error("{} not found".format(boid_or_localobject))
        else:
            return item



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Write.py - creation of WriteProperty requests

    Used while defining an app
    Example::

        class BasicScript(WhoisIAm, WriteProperty)

    Class::

        WriteProperty()
            def write()


"""
# --- 3rd party modules ---
from bacpypes.debugging import bacpypes_debugging, ModuleLogger

from bacpypes.pdu import Address
from bacpypes.object import get_datatype

from bacpypes.apdu import (
    WritePropertyRequest,
    WritePropertyMultipleRequest,
    SimpleAckPDU,
    WriteAccessSpecification,
)

from bacpypes.primitivedata import Null, Atomic, Integer, Unsigned, Real, Enumerated
from bacpypes.constructeddata import Array, Any, SequenceOf
from bacpypes.basetypes import PropertyValue, PropertyIdentifier
from bacpypes.iocb import IOCB
from bacpypes.core import deferred

# --- this application's modules ---
from .IOExceptions import (
    WritePropertyCastError,
    NoResponseFromController,
    WritePropertyException,
    WriteAccessDenied,
    ApplicationNotStarted,
)
from ...core.utils.notes import note_and_log
from .Read import find_reason

# ------------------------------------------------------------------------------

# some debugging
_debug = 0
_LOG = ModuleLogger(globals())


@note_and_log
class WriteProperty:
    """
    Defines BACnet Write functions: WriteProperty [WritePropertyMultiple not supported]

    """

    def write(self, args, vendor_id=0, timeout=10):
        """ Build a WriteProperty request, wait for an answer, and return status [True if ok, False if not].

        :param args: String with <addr> <type> <inst> <prop> <value> [ <indx> ] - [ <priority> ]
        :returns: return status [True if ok, False if not]

        *Example*::

            import BAC0
            bacnet = BAC0.lite()
            bacnet.write('2:5 analogValue 1 presentValue 100 - 8')

        Direct the controller at (Network 2, address 5) to write 100 to the presentValues of
        its analogValue 1 (AV:1) at priority 8
        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")
        args = args.split()
        self.log_title("Write property", args)

        try:
            # build a WriteProperty request
            iocb = IOCB(self.build_wp_request(args, vendor_id=vendor_id))
            iocb.set_timeout(timeout)
            # pass to the BACnet stack
            deferred(self.this_application.request_io, iocb)
            self._log.debug("{:<20} {!r}".format("iocb", iocb))

        except WritePropertyException as error:
            # construction error
            self._log.exception("exception: {!r}".format(error))

        iocb.wait()  # Wait for BACnet response

        if iocb.ioResponse:  # successful response
            apdu = iocb.ioResponse

            if not isinstance(apdu, SimpleAckPDU):  # expect an ACK
                self._log.warning("Not an ack, see debug for more infos.")
                self._log.debug(
                    "Not an ack. | APDU : {} / {}".format((apdu, type(apdu)))
                )
                return

        if iocb.ioError:  # unsuccessful: error/reject/abort
            apdu = iocb.ioError
            reason = find_reason(apdu)
            raise NoResponseFromController("APDU Abort Reason : {}".format(reason))

    def _parse_wp_args(self, args):
        """
        Utility to parse the string of the request.
        Supports @obj_ and @prop_ syntax for objest type and property id, useful with proprietary objects and properties.
        """
        if not isinstance(args, list):
            args = args.split()
        obj_type, obj_inst, prop_id = args[:3]
        if obj_type.isdigit():
            obj_type = int(obj_type)
        elif "@obj_" in obj_type:
            obj_type = int(obj_type.split("_")[1])
        obj_inst = int(obj_inst)
        value = args[3]
        indx = None
        if len(args) >= 5 and args[4] != "-":
            indx = int(args[4])
        priority = None
        if len(args) >= 6:
            priority = int(args[5])
        if "@prop_" in prop_id:
            prop_id = prop_id.split("_")[1]
        if prop_id.isdigit():
            prop_id = int(prop_id)

        return (obj_type, obj_inst, prop_id, value, priority, indx)

    def _validate_value_vs_datatype(self, obj_type, prop_id, indx, vendor_id, value):
        """
        This will ensure the value can be encoded and is valid in the context
        """
        # get the datatype
        datatype = get_datatype(obj_type, prop_id, vendor_id=vendor_id)
        # change atomic values into something encodeable, null is a special
        # case
        if value == "null":
            value = Null()
        elif issubclass(datatype, Atomic):
            if (
                datatype is Integer
                # or datatype is not Real
                or datatype is Unsigned
                or datatype is Enumerated
            ):
                value = int(value)
            elif datatype is Real:
                value = float(value)
                # value = datatype(value)
            else:
                # value = float(value)
                value = datatype(value)

            value = datatype(value)
        elif issubclass(datatype, Array) and (indx is not None):
            if indx == 0:
                value = Integer(value)
            elif issubclass(datatype.subtype, Atomic):
                value = datatype.subtype(value)
            elif not isinstance(value, datatype.subtype):
                raise TypeError(
                    "invalid result datatype, expecting {}".format(
                        (datatype.subtype.__name__,)
                    )
                )

        elif not isinstance(value, datatype):
            raise TypeError(
                "invalid result datatype, expecting {}".format((datatype.__name__,))
            )

        self._log.debug("{:<20} {!r} {}".format("Encodeable value", value, type(value)))

        _value = Any()
        try:
            _value.cast_in(value)
        except WritePropertyCastError as error:
            self._log.error("WriteProperty cast error: {!r}".format(error))
            raise

        return _value

    def build_wp_request(self, args, vendor_id=0):
        vendor_id = vendor_id
        addr = args[0]
        args = args[1:]
        obj_type, obj_inst, prop_id, value, priority, indx = self._parse_wp_args(args)

        value = self._validate_value_vs_datatype(
            obj_type, prop_id, indx, vendor_id, value
        )
        # build a request
        request = WritePropertyRequest(
            objectIdentifier=(obj_type, obj_inst), propertyIdentifier=prop_id
        )
        request.pduDestination = Address(addr)

        # save the value
        request.propertyValue = value

        # optional array index
        if indx is not None:
            request.propertyArrayIndex = indx

        # optional priority
        if priority is not None:
            request.priority = priority

        self.log_subtitle("Creating Request")
        self._log.debug(
            "{:<20} {:<20} {:<20} {:<20}".format(
                "indx", "priority", "datatype", "value"
            )
        )
        datatype = get_datatype(obj_type, prop_id, vendor_id=vendor_id)

        self._log.debug(
            "{!r:<20} {!r:<20} {!r:<20} {!r:<20}".format(
                indx, priority, datatype, value
            )
        )

        self._log.debug("{:<20} {}".format("REQUEST", request))
        return request

    def writeMultiple(self, addr=None, args=None, vendor_id=0, timeout=10):
        """ Build a WritePropertyMultiple request, wait for an answer

        :param addr: destination of request (ex. '2:3' or '192.168.1.2')
        :param args: list of String with <type> <inst> <prop> <value> [ <indx> ] - [ <priority> ]
        :param vendor_id: Mandatory for registered proprietary object and properties
        :param timeout: used by IOCB to discard request if timeout reached
        :returns: return status [True if ok, False if not]

        *Example*::

            import BAC0
            bacnet = BAC0.lite()
            r = ['analogValue 1 presentValue 100','analogValue 2 presentValue 100','analogValue 3 presentValue 100 - 8','@obj_142 1 @prop_1042 True']
            bacnet.writeMultiple(addr='2:5',args=r,vendor_id=842)
            # or 
            # bacnet.writeMultiple('2:5',r)
            
        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        self.log_title("Write property multiple", args)

        try:
            # build a WritePropertyMultiple request
            iocb = IOCB(self.build_wpm_request(args, vendor_id=vendor_id, addr=addr))
            iocb.set_timeout(timeout)
            # pass to the BACnet stack
            deferred(self.this_application.request_io, iocb)
            self._log.debug("{:<20} {!r}".format("iocb", iocb))

        except WritePropertyException as error:
            # construction error
            self._log.exception("exception: {!r}".format(error))

        iocb.wait()  # Wait for BACnet response

        if iocb.ioResponse:  # successful response
            apdu = iocb.ioResponse

            if not isinstance(apdu, SimpleAckPDU):  # expect an ACK
                self._log.warning("Not an ack, see debug for more infos.")
                self._log.debug(
                    "Not an ack. | APDU : {} / {}".format((apdu, type(apdu)))
                )
                return

        if iocb.ioError:  # unsuccessful: error/reject/abort
            apdu = iocb.ioError
            reason = find_reason(apdu)
            raise NoResponseFromController("APDU Abort Reason : {}".format(reason))

    def build_wpm_request(self, args, vendor_id=0, addr=None):
        if not addr:
            raise ValueError("Please provide addr")

        self.log_subtitle("Creating Write Multiple Request")
        self._log.debug(
            "{:<20} {:<20} {:<20} {:<20}".format(
                "indx", "priority", "datatype", "value"
            )
        )

        was = []
        for each in args:
            property_values = []
            if isinstance(each, str):
                obj_type, obj_inst, prop_id, value, priority, indx = self._parse_wp_args(
                    each
                )
            elif isinstance(each, tuple):
                # Supported but not really the best choice as the parser will
                # catch some edge cases for you
                obj_type, obj_inst, prop_id, value, priority, indx = each
            else:
                raise ValueError("Wrong encoding of request")
            value = self._validate_value_vs_datatype(
                obj_type, prop_id, indx, vendor_id, value
            )

            existingObject = next(
                (obj for obj in was if obj.objectIdentifier == (obj_type, obj_inst)),
                None,
            )

            if existingObject == None:
                property_values.append(
                    PropertyValue(
                        propertyIdentifier=prop_id,
                        propertyArrayIndex=indx,
                        value=value,
                        priority=priority,
                    )
                )

                was.append(
                    WriteAccessSpecification(
                        objectIdentifier=(obj_type, obj_inst),
                        listOfProperties=property_values,
                    )
                )
            else:
                existingObject.listOfProperties.append(
                    PropertyValue(
                        propertyIdentifier=prop_id,
                        propertyArrayIndex=indx,
                        value=value,
                        priority=priority,
                    )
                )

            datatype = get_datatype(obj_type, prop_id, vendor_id=vendor_id)

            self._log.debug(
                "{!r:<20} {!r:<20} {!r:<20} {!r:<20}".format(
                    indx, priority, datatype, value
                )
            )

        # build a request
        request = WritePropertyMultipleRequest(listOfWriteAccessSpecs=was)
        request.pduDestination = Address(addr)

        self._log.debug("{:<20} {}".format("REQUEST", request))
        return request



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Discover.py

Classes needed to make discovering functions on a BACnet network

"""
# --- standard Python modules ---
import time

# --- 3rd party modules ---
from bacpypes.apdu import (
    WhoIsRequest,
    IAmRequest,
    WhoHasRequest,
    WhoHasLimits,
    WhoHasObject,
)
from bacpypes.core import deferred
from bacpypes.pdu import Address, GlobalBroadcast, LocalBroadcast
from bacpypes.primitivedata import Unsigned, ObjectIdentifier, CharacterString
from bacpypes.constructeddata import Array
from bacpypes.object import get_object_class, get_datatype
from bacpypes.iocb import IOCB, SieveQueue, IOController

from bacpypes.netservice import NetworkServiceAccessPoint, NetworkServiceElement
from bacpypes.npdu import (
    WhoIsRouterToNetwork,
    IAmRouterToNetwork,
    InitializeRoutingTable,
    InitializeRoutingTableAck,
    WhatIsNetworkNumber,
    NetworkNumberIs,
    RejectMessageToNetwork,
)

# --- this application's modules ---
from ..io.IOExceptions import (
    SegmentationNotSupported,
    ReadPropertyException,
    ReadPropertyMultipleException,
    NoResponseFromController,
    ApplicationNotStarted,
)
from ...core.utils.notes import note_and_log

# ------------------------------------------------------------------------------
@note_and_log
class NetworkServiceElementWithRequests(IOController, NetworkServiceElement):
    """
    This class will add the capability to send requests at network level
    And capability to read responses for NPDU
    Deals with IOCB so the request can be deferred to task manager

    """

    def __init__(self):
        NetworkServiceElement.__init__(self)
        IOController.__init__(self)

        # no pending request
        self._request = None
        self._iartn = []
        self._learnedNetworks = set()
        self.queue_by_address = {}
        self._routing_table = {}

    def process_io(self, iocb):
        # get the destination address from the pdu
        adapter, npdu = iocb.args[0]
        destination_address = npdu.pduDestination

        # look up the queue
        queue = self.queue_by_address.get(destination_address, None)
        if not queue:
            queue = SieveQueue(self.request, address=destination_address)
            self.queue_by_address[destination_address] = queue
        # ask the queue to process the request
        queue.request_io(iocb)

    def _net_complete(self, npdu):

        # look up the queue
        queue = self.queue_by_address.get(npdu.pduDestination, None)
        if not queue:
            return

        # make sure it has an active iocb
        if not queue.active_iocb:
            return

        # this request is complete
        if isinstance(
            npdu,
            (
                None.__class__,
                IAmRouterToNetwork,
                InitializeRoutingTableAck,
                NetworkNumberIs,
            ),
        ):
            queue.complete_io(queue.active_iocb, npdu)
        elif isinstance(npdu, RejectMessageToNetwork):
            queue.abort_io(queue.active_iocb, npdu)
        else:
            raise RuntimeError("unrecognized NPDU type")

        # if the queue is empty and idle, forget about the controller
        if not queue.ioQueue.queue and not queue.active_iocb:
            del self.queue_by_address[npdu.pduDestination]

    def request(self, arg):
        adapter, npdu = arg
        # save a copy of the request
        self._request = npdu

        # forward it along
        NetworkServiceElement.request(self, adapter, npdu)

    def indication(self, adapter, npdu):
        if isinstance(npdu, IAmRouterToNetwork):
            if isinstance(self._request, WhoIsRouterToNetwork):
                address, netlist = str(npdu.pduSource), npdu.iartnNetworkList
                self._log.info("{} router to {}".format(address, netlist))
                self._iartn.append(address)
                self._routing_table[address] = netlist
            for each in npdu.iartnNetworkList:
                self._learnedNetworks.add(int(each))

        elif isinstance(npdu, InitializeRoutingTableAck):
            self._log.info("{} routing table".format(npdu.pduSource))
            for rte in npdu.irtaTable:
                self._log.info(
                    "    {} {} {}".format(rte.rtDNET, rte.rtPortID, rte.rtPortInfo)
                )

        elif isinstance(npdu, NetworkNumberIs):
            self._log.info(
                "{} network number is {}".format(npdu.pduSource, npdu.nniNet)
            )
            self._learnedNetworks.add(int(npdu.nniNet))

        elif isinstance(npdu, RejectMessageToNetwork):
            self._log.warning(
                "{} Rejected message to network (reason : {})".format(
                    npdu.pduSource,
                    rejectMessageToNetworkReasons[npdu.rmtnRejectionReason],
                )
            )
        # forward it along
        NetworkServiceElement.indication(self, adapter, npdu)

    def response(self, adapter, npdu):
        # forward it along
        NetworkServiceElement.response(self, adapter, npdu)

    def confirmation(self, adapter, npdu):
        # forward it along
        self._net_complete(npdu)
        NetworkServiceElement.confirmation(self, adapter, npdu)


@note_and_log
class Discover:
    """
    Define BACnet WhoIs and IAm functions.
    """

    def whois(self, *args, global_broadcast=False, destination=None):
        """
        Build a WhoIs request

        :param args: string built as [ <addr>] [ <lolimit> <hilimit> ] **optional**
        :returns: discoveredDevices as a defaultdict(int)

        Example::

            whois(global_broadcast=True) # WhoIs broadcast globally.  Every device will respond with an IAm
            whois('2:5')                 # WhoIs looking for the device at (Network 2, Address 5)
            whois('10 1000')             # WhoIs looking for devices in the ID range (10 - 1000)

        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        if args:
            args = args[0].split()
        msg = args if args else "any"

        self._log.debug("do_whois {!r}".format(msg))

        # build a request
        request = WhoIsRequest()
        if (len(args) == 1) or (len(args) == 3):
            request.pduDestination = Address(args[0])
            del args[0]
        else:
            if global_broadcast:
                request.pduDestination = GlobalBroadcast()
            else:
                request.pduDestination = LocalBroadcast()

        if len(args) == 2:
            try:
                request.deviceInstanceRangeLowLimit = int(args[0])
                request.deviceInstanceRangeHighLimit = int(args[1])
            except ValueError:
                pass
        self._log.debug("{:>12} {}".format("- request:", request))

        if destination:
            request.pduDestination = Address(destination)
        iocb = IOCB(request)  # make an IOCB
        self.this_application._last_i_am_received = []
        # pass to the BACnet stack
        deferred(self.this_application.request_io, iocb)

        iocb.wait()  # Wait for BACnet response

        if iocb.ioResponse:  # successful response
            apdu = iocb.ioResponse

        if iocb.ioError:  # unsuccessful: error/reject/abort
            pass

        time.sleep(3)
        self.discoveredDevices = self.this_application.i_am_counter
        return self.this_application._last_i_am_received

    def _iam_request(self, destination=None):
        """
        Build the IOCB request for a I Am
        """
        try:
            # build a response
            request = IAmRequest()
            request.pduDestination = destination if destination else GlobalBroadcast()
            # fill the response with details about us (from our device object)
            request.iAmDeviceIdentifier = self.this_device.objectIdentifier
            request.maxAPDULengthAccepted = self.this_device.maxApduLengthAccepted
            request.segmentationSupported = self.this_device.segmentationSupported
            request.vendorID = self.this_device.vendorIdentifier
            self._log.debug("{:>12} {}".format("- request:", request))

            return request

        except Exception as error:
            self._log.error("exception: {!r}".format(error))
            raise

    def iam(self, destination=None):
        """
        Build an IAm response.  IAm are sent in response to a WhoIs request that;
        matches our device ID, whose device range includes us, or is a broadcast.
        Content is defined by the script (deviceId, vendor, etc...)

        :returns: bool

        Example::

            iam()
        """

        self._log.debug("do_iam")

        try:
            # build a response
            request = self._iam_request(destination=destination)
            iocb = IOCB(request)  # make an IOCB
            deferred(self.this_application.request_io, iocb)
            iocb.wait()
            return True

        except Exception as error:
            self._log.error("exception: {!r}".format(error))
            return False

    def whois_router_to_network(self, network=None, *, destination=None):
        # build a request
        try:
            request = WhoIsRouterToNetwork()
            if network:
                request.wirtnNetwork = int(network)
            if destination:
                request.pduDestination = Address(destination)
                self._log.debug(
                    "WhoIsRouterToNetwork Destination : {}".format(destination)
                )
            else:
                request.pduDestination = LocalBroadcast()
        except:
            self._log.error("WhoIsRouterToNetwork : invalid arguments")
            return
        iocb = IOCB((self.this_application.nsap.local_adapter, request))  # make an IOCB
        iocb.set_timeout(2)
        deferred(self.this_application.nse.request_io, iocb)
        iocb.wait()

        try:
            self.init_routing_table(str(self.this_application.nse._iartn.pop()))
        except IndexError:
            pass

    def init_routing_table(self, address):
        """
        irt <addr>

        Send an empty Initialize-Routing-Table message to an address, a router
        will return an acknowledgement with its routing table configuration.
        """
        # build a request
        self._log.info("Addr : {}".format(address))
        try:
            request = InitializeRoutingTable()
            request.pduDestination = Address(address)
        except:
            self._log.error("invalid arguments")
            return

        iocb = IOCB((self.this_application.nsap.local_adapter, request))  # make an IOCB
        iocb.set_timeout(2)
        deferred(self.this_application.nse.request_io, iocb)
        iocb.wait()

    def what_is_network_number(self, destination=None):
        """
        winn [ <addr> ]

        Send a What-Is-Network-Number message.  If the address is unspecified
        the message is locally broadcast.
        """
        # build a request

        request = WhatIsNetworkNumber()
        if destination:
            request.pduDestination = Address(destination)
        else:
            request.pduDestination = LocalBroadcast()

        iocb = IOCB((self.this_application.nsap.local_adapter, request))  # make an IOCB
        iocb.set_timeout(2)
        deferred(self.this_application.nse.request_io, iocb)
        iocb.wait()

    def whohas(
        self,
        object_id=None,
        object_name=None,
        instance_range_low_limit=0,
        instance_range_high_limit=4194303,
        destination=None,
        global_broadcast=False,
    ):
        """
        Object ID : analogInput:1
        Object Name : string
        Instance Range Low Limit : 0
        Instance Range High Limit : 4194303
        destination (optional) : If empty, local broadcast will be used.
        global_broadcast : False

        """
        obj_id = ObjectIdentifier(object_id)
        if object_name and not object_id:
            obj_name = CharacterString(object_name)
            obj = WhoHasObject(objectName=obj_name)
        elif object_id and not object_name:
            obj = WhoHasObject(objectIdentifier=obj_id)
        else:
            obj = WhoHasObject(objectIdentifier=obj_id, objectName=obj_name)
        limits = WhoHasLimits(
            deviceInstanceRangeLowLimit=instance_range_low_limit,
            deviceInstanceRangeHighLimit=instance_range_high_limit,
        )
        request = WhoHasRequest(object=obj, limits=limits)
        if destination:
            request.pduDestination = Address(destination)
        else:
            if global_broadcast:
                request.pduDestination = GlobalBroadcast()
            else:
                request.pduDestination = LocalBroadcast()
        iocb = IOCB(request)  # make an IOCB
        iocb.set_timeout(2)
        deferred(self.this_application.request_io, iocb)
        iocb.wait()

        iocb = IOCB(request)  # make an IOCB
        self.this_application._last_i_have_received = []

        if iocb.ioResponse:  # successful response
            apdu = iocb.ioResponse

        if iocb.ioError:  # unsuccessful: error/reject/abort
            pass

        time.sleep(3)
        # self.discoveredObjects = self.this_application.i_am_counter
        return self.this_application._last_i_have_received


rejectMessageToNetworkReasons = [
    "Other Error",
    "The router is not direclty connected to DNET and cannot find a router to DNET on any direclty connected network using Who-Is-Router-To-Network messages",
    "The tour is busy and unable to accept messages for the specified DNET at the present time",
    "It is an unknown network layer message",
    "The message is too long to be routed to this DNET",
    "The source message was rejected due to a BACnet security error and that error cannot be forwarded to the source device",
    "The source message was rejected due to errors in the addressing. The length of th DADR or SADR was determined to be invalid",
]



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
TaskManager.py - creation of threads used for repetitive tasks.  

A key building block for point simulation.
"""
# --- standard Python modules ---
from threading import Thread, Lock
import time

# --- 3rd party modules ---
# --- this application's modules ---
from ..core.utils.notes import note_and_log

# ------------------------------------------------------------------------------


class Manager:
    taskList = []
    threadLock = Lock()


def stopAllTasks():
    for each in Manager.taskList:
        each.exitFlag = True
    while True:
        _alive = [each.is_alive() for each in Manager.taskList]
        if not any(_alive):
            clean_tasklist()
            break
    return True


def clean_tasklist():
    for each in Manager.taskList:
        if not each.is_alive():
            Manager.taskList.remove(each)


@note_and_log
class Task(Thread):
    def __init__(self, delay=5, daemon=True, name="recurring"):
        Thread.__init__(self, name=name, daemon=daemon)
        self.is_running = False
        self.exitFlag = False
        self.lock = Manager.threadLock
        self.delay = delay
        if self.name not in Manager.taskList:
            Manager.taskList.append(self)

    def run(self):
        self.process()

    def process(self):
        self.is_running = True
        while not self.exitFlag:
            self.lock.acquire()
            self.task()
            self.lock.release()
            # This replace a single time.sleep
            # the goal is to speed up the stop
            # of the thread by providing an easy way out
            for _ in range(self.delay * 2):
                if self.exitFlag:
                    break
                time.sleep(0.5)
            clean_tasklist()

    def task(self):
        raise RuntimeError("task must be overridden")

    def stop(self):
        self.is_running = False
        self.exitFlag = True
        clean_tasklist()

    def beforeStop(self):
        """
        Action done when closing thread
        """
        if self in Manager.taskList:
            Manager.taskList.remove(self)


@note_and_log
class OneShotTask(Thread):
    def __init__(self, daemon=True, name="Oneshot"):
        Thread.__init__(self, name=name, daemon=daemon)
        self.lock = Manager.threadLock
        if self.name not in Manager.taskList:
            Manager.taskList.append(self)

    def run(self):
        self.process()

    def process(self):
        self.task()

    def task(self):
        raise RuntimeError("task must be overridden")

    def stop(self):
        pass

    def beforeStop(self):
        """
        Action done when closing thread
        """
        if self in Manager.taskList:
            Manager.taskList.remove(self)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Read.py - creation of ReadProperty and ReadPropertyMultiple requests

    Used while defining an app:
    Example::

        class BasicScript(WhoisIAm, ReadProperty)

    Class::

        ReadProperty()
            def read()
            def readMultiple()

"""

# --- standard Python modules ---

# --- 3rd party modules ---
from bacpypes.debugging import bacpypes_debugging

from bacpypes.pdu import Address
from bacpypes.object import get_object_class, get_datatype
from bacpypes.apdu import (
    PropertyReference,
    ReadAccessSpecification,
    ReadPropertyRequest,
    ReadPropertyMultipleRequest,
    RejectReason,
    AbortReason,
    RejectPDU,
    AbortPDU,
)

from bacpypes.basetypes import PropertyIdentifier, DateTime
from bacpypes.apdu import (
    ReadPropertyMultipleACK,
    ReadPropertyACK,
    ReadRangeRequest,
    ReadRangeACK,
    Range,
    RangeByPosition,
    RangeBySequenceNumber,
    RangeByTime,
)
from bacpypes.primitivedata import Tag, ObjectIdentifier, Unsigned, Date, Time
from bacpypes.constructeddata import Array
from bacpypes.iocb import IOCB, TimeoutError
from bacpypes.core import deferred

# --- this application's modules ---
from .IOExceptions import (
    ReadPropertyException,
    ReadPropertyMultipleException,
    ReadRangeException,
    NoResponseFromController,
    ApplicationNotStarted,
    UnrecognizedService,
    SegmentationNotSupported,
    UnknownPropertyError,
    UnknownObjectError,
    BufferOverflow,
)

from ..utils.notes import note_and_log

# ------------------------------------------------------------------------------


@note_and_log
class ReadProperty:
    """
    Defines BACnet Read functions: readProperty and readPropertyMultiple.
    Data exchange is made via a Queue object
    A timeout of 10 seconds allows detection of invalid device or communciation errors.
    """

    def read(
        self,
        args,
        arr_index=None,
        vendor_id=0,
        bacoid=None,
        timeout=10,
        prop_id_required=False,
    ):
        """
        Build a ReadProperty request, wait for the answer and return the value

        :param args: String with <addr> <type> <inst> <prop> [ <indx> ]
        :returns: data read from device (str representing data like 10 or True)

        *Example*::

            import BAC0
            myIPAddr = '192.168.1.10/24'
            bacnet = BAC0.connect(ip = myIPAddr)
            bacnet.read('2:5 analogInput 1 presentValue')

        Requests the controller at (Network 2, address 5) for the presentValue of
        its analog input 1 (AI:1).
        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        args_split = args.split()

        self.log_title("Read property", args_split)

        vendor_id = vendor_id
        bacoid = bacoid

        try:
            # build ReadProperty request
            iocb = IOCB(
                self.build_rp_request(
                    args_split, arr_index=arr_index, vendor_id=vendor_id, bacoid=bacoid
                )
            )
            iocb.set_timeout(timeout)
            # pass to the BACnet stack
            deferred(self.this_application.request_io, iocb)
            self._log.debug("{:<20} {!r}".format("iocb", iocb))

        except ReadPropertyException as error:
            # construction error
            self._log.exception("exception: {!r}".format(error))

        iocb.wait()  # Wait for BACnet response

        if iocb.ioResponse:  # successful response
            apdu = iocb.ioResponse

            if not isinstance(apdu, ReadPropertyACK):  # expecting an ACK
                self._log.warning("Not an ack, see debug for more infos.")
                self._log.debug("Not an ack. | APDU : {}".format(apdu))
                return

            # find the datatype
            datatype = get_datatype(
                apdu.objectIdentifier[0], apdu.propertyIdentifier, vendor_id=vendor_id
            )
            if not datatype:
                # raise TypeError("unknown datatype")
                value = cast_datatype_from_tag(
                    apdu.propertyValue,
                    apdu.objectIdentifier[0],
                    apdu.propertyIdentifier,
                )
            else:
                # special case for array parts, others are managed by cast_out
                if issubclass(datatype, Array) and (
                    apdu.propertyArrayIndex is not None
                ):
                    if apdu.propertyArrayIndex == 0:
                        value = apdu.propertyValue.cast_out(Unsigned)
                    else:
                        value = apdu.propertyValue.cast_out(datatype.subtype)
                else:
                    value = apdu.propertyValue.cast_out(datatype)

                self._log.debug("{:<20} {:<20}".format("value", "datatype"))
                self._log.debug("{!r:<20} {!r:<20}".format(value, datatype))
            if prop_id_required:
                try:
                    int(apdu.propertyIdentifier)
                    prop_id = "@prop_{}".format(apdu.propertyIdentifier)
                    value = list(value.items())[0][1]
                except ValueError:
                    prop_id = apdu.propertyIdentifier
                return (value, prop_id)
            else:
                return value

        if iocb.ioError:  # unsuccessful: error/reject/abort
            apdu = iocb.ioError
            reason = find_reason(apdu)
            if reason == "segmentationNotSupported":
                value = self._split_the_read_request(args, arr_index)
                return value
            else:
                if reason == "unknownProperty":
                    if "description" in args:
                        self._log.warning(
                            "The description property is not implemented in the device. Using a default value for internal needs."
                        )
                        return "Property Not Implemented"
                    elif "inactiveText" in args:
                        self._log.warning(
                            "The inactiveText property is not implemented in the device. Using a default value of Off for internal needs."
                        )
                        return "Off"
                    elif "activeText" in args:
                        self._log.warning(
                            "The activeText property is not implemented in the device. Using a default value of On for internal needs."
                        )
                        return "On"
                    else:
                        raise UnknownPropertyError("Unknown property {}".format(args))
                elif reason == "unknownObject":
                    self._log.warning("Unknown object {}".format(args))
                    raise UnknownObjectError("Unknown object {}".format(args))
                elif reason == "bufferOverflow":
                    self._log.warning(
                        "Buffer capacity exceeded in device {}".format(args)
                    )
                    return self._split_the_read_request(args, arr_index)
                else:
                    # Other error... consider NoResponseFromController (65)
                    # even if the real reason is another one
                    raise NoResponseFromController(
                        "APDU Abort Reason : {}".format(reason)
                    )

    def _split_the_read_request(self, args, arr_index):
        """
        When a device doesn't support segmentation, this function
        will split the request according to the length of the
        predicted result which can be known when reading the array_index
        number 0.

        This can be a very long process as some devices count a large
        number of properties without supporting segmentation
        (FieldServers are a good example)
        """
        nmbr_obj = self.read(args, arr_index=0)
        return [self.read(args, arr_index=i) for i in range(1, nmbr_obj + 1)]

    def readMultiple(
        self, args, request_dict=None, vendor_id=0, timeout=10, prop_id_required=False
    ):
        """ Build a ReadPropertyMultiple request, wait for the answer and return the values

        :param args: String with <addr> ( <type> <inst> ( <prop> [ <indx> ] )... )...
        :returns: data read from device (str representing data like 10 or True)

        *Example*::

            import BAC0
            myIPAddr = '192.168.1.10/24'
            bacnet = BAC0.connect(ip = myIPAddr)
            bacnet.readMultiple('2:5 analogInput 1 presentValue units')

        Requests the controller at (Network 2, address 5) for the (presentValue and units) of
        its analog input 1 (AI:1).
        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        if request_dict is not None:
            request = self.build_rpm_request_from_dict(request_dict, vendor_id)
        else:
            args = args.split()
            request = self.build_rpm_request(args, vendor_id=vendor_id)
            self.log_title("Read Multiple", args)

        values = []
        dict_values = {}

        try:
            # build an ReadPropertyMultiple request
            iocb = IOCB(request)
            iocb.set_timeout(timeout)
            # pass to the BACnet stack
            deferred(self.this_application.request_io, iocb)
            self._log.debug("{:<20} {!r}".format("iocb", iocb))

        except ReadPropertyMultipleException as error:
            # construction error
            self._log.exception("exception: {!r}".format(error))

        iocb.wait()  # Wait for BACnet response

        if iocb.ioResponse:  # successful response
            apdu = iocb.ioResponse

            if not isinstance(apdu, ReadPropertyMultipleACK):  # expecting an ACK
                self._log.debug("{:<20}".format("not an ack"))
                self._log.warning(
                    "Not an Ack. | APDU : {} / {}".format((apdu, type(apdu)))
                )
                return

            # loop through the results
            for result in apdu.listOfReadAccessResults:
                # here is the object identifier
                objectIdentifier = result.objectIdentifier

                self.log_subtitle(
                    "{!r} : {!r}".format(objectIdentifier[0], objectIdentifier[1]),
                    width=114,
                )
                self._log.debug(
                    "{:<20} {:<20} {:<30} {:<20}".format(
                        "propertyIdentifier", "propertyArrayIndex", "value", "datatype"
                    )
                )
                self._log.debug("-" * 114)
                dict_values[objectIdentifier] = []
                # now come the property values per object
                for element in result.listOfResults:
                    # get the property and array index
                    propertyIdentifier = element.propertyIdentifier
                    propertyArrayIndex = element.propertyArrayIndex

                    readResult = element.readResult

                    if propertyArrayIndex is not None:
                        _prop_id = "{}@idx:{}".format(
                            propertyIdentifier, propertyArrayIndex
                        )
                    else:
                        _prop_id = propertyIdentifier

                    if readResult.propertyAccessError is not None:
                        self._log.debug(
                            "Property Access Error for {}".format(
                                readResult.propertyAccessError
                            )
                        )
                        values.append(None)
                        dict_values[objectIdentifier].append((_prop_id, None))
                    else:
                        # here is the value
                        propertyValue = readResult.propertyValue

                        # find the datatype
                        datatype = get_datatype(
                            objectIdentifier[0], propertyIdentifier, vendor_id=vendor_id
                        )

                        if not datatype:
                            value = cast_datatype_from_tag(
                                propertyValue, objectIdentifier[0], propertyIdentifier
                            )
                        else:
                            # special case for array parts, others are managed by cast_out
                            if issubclass(datatype, Array) and (
                                propertyArrayIndex is not None
                            ):
                                if propertyArrayIndex == 0:
                                    value = propertyValue.cast_out(Unsigned)
                                else:
                                    value = propertyValue.cast_out(datatype.subtype)
                            elif propertyValue.is_application_class_null():
                                value = None
                            else:
                                value = propertyValue.cast_out(datatype)

                            self._log.debug(
                                "{!r:<20} {!r:<20} {!r:<30} {!r:<20}".format(
                                    propertyIdentifier,
                                    propertyArrayIndex,
                                    value,
                                    datatype,
                                )
                            )
                        if prop_id_required:
                            try:
                                int(propertyIdentifier)
                                prop_id = "@prop_{}".format(propertyIdentifier)
                                value = list(value.items())[0][1]
                            except ValueError:
                                prop_id = propertyIdentifier
                            values.append((value, prop_id))
                            dict_values[objectIdentifier].append(
                                (_prop_id, (value, prop_id))
                            )
                        else:
                            values.append(value)
                            dict_values[objectIdentifier].append((_prop_id, value))

            if request_dict is not None:
                return dict_values
            else:
                return values

        if iocb.ioError:  # unsuccessful: error/reject/abort
            apdu = iocb.ioError
            reason = find_reason(apdu)
            self._log.warning("APDU Abort Reject Reason : {}".format(reason))
            self._log.debug("The Request was : {}".format(args))
            if reason == "unrecognizedService":
                raise UnrecognizedService()
            elif reason == "segmentationNotSupported":
                # value = self._split_the_read_request(args, arr_index)
                # return value
                self.segmentation_supported = False
                raise SegmentationNotSupported()
            elif reason == "unknownObject":
                self._log.warning("Unknown object {}".format(args))
                raise UnknownObjectError("Unknown object {}".format(args))
            elif reason == "unknownProperty":
                self._log.warning("Unknown property {}".format(args))

                values.append("")
                return values
            else:
                self._log.warning("No response from controller {}".format(reason))
                values.append("")
                return values

    def build_rp_request(self, args, arr_index=None, vendor_id=0, bacoid=None):
        addr, obj_type, obj_inst, prop_id = args[:4]
        vendor_id = vendor_id
        bacoid = bacoid

        if obj_type.isdigit():
            obj_type = int(obj_type)
        elif "@obj_" in obj_type:
            obj_type = int(obj_type.split("_")[1])
        elif not get_object_class(obj_type, vendor_id=vendor_id):
            raise ValueError("Unknown object type : {}".format(obj_type))

        obj_inst = int(obj_inst)

        if prop_id.isdigit():
            prop_id = int(prop_id)
        elif "@prop_" in prop_id:
            prop_id = int(prop_id.split("_")[1])

        # datatype = get_datatype(obj_type, prop_id, vendor_id=vendor_id)

        # build a request
        request = ReadPropertyRequest(
            objectIdentifier=(obj_type, obj_inst),
            propertyIdentifier=prop_id,
            propertyArrayIndex=arr_index,
        )
        request.pduDestination = Address(addr)

        if len(args) == 5:
            request.propertyArrayIndex = int(args[4])
        self._log.debug("{:<20} {!r}".format("REQUEST", request))
        return request

    def build_rpm_request(self, args, vendor_id=0):
        """
        Build request from args
        """
        self._log.debug(args)
        i = 0
        addr = args[i]
        i += 1
        vendor_id = vendor_id

        read_access_spec_list = []
        while i < len(args):
            obj_type = args[i]
            i += 1

            if obj_type.isdigit():
                obj_type = int(obj_type)
            elif "@obj_" in obj_type:
                obj_type = int(obj_type.split("_")[1])
            elif not get_object_class(obj_type, vendor_id=vendor_id):
                raise ValueError("Unknown object type : {}".format(obj_type))

            obj_inst = int(args[i])
            i += 1

            prop_reference_list = []
            while i < len(args):
                prop_id = args[i]
                if "@obj_" in prop_id:
                    break
                if prop_id not in PropertyIdentifier.enumerations:
                    try:
                        if "@prop_" in prop_id:
                            prop_id = int(prop_id.split("_")[1])
                            self._log.debug(
                                "Proprietary property : {} | {} -> Vendor : {}".format(
                                    obj_type, prop_id, vendor_id
                                )
                            )
                        else:
                            break
                    except:
                        break

                elif prop_id not in (
                    "all",
                    "required",
                    "optional",
                    "objectName",
                    "objectType",
                    "objectIdentifier",
                    "polarity",
                ):
                    datatype = get_datatype(obj_type, prop_id, vendor_id=vendor_id)
                    if not datatype:
                        raise ValueError(
                            "invalid property for object type : {} | {}".format(
                                obj_type, prop_id
                            )
                        )
                i += 1

                # build a property reference
                prop_reference = PropertyReference(propertyIdentifier=prop_id)

                # check for an array index
                if (i < len(args)) and args[i].isdigit():
                    prop_reference.propertyArrayIndex = int(args[i])
                    i += 1

                prop_reference_list.append(prop_reference)

            if not prop_reference_list:
                raise ValueError("provide at least one property")

            # build a read access specification
            read_access_spec = ReadAccessSpecification(
                objectIdentifier=(obj_type, obj_inst),
                listOfPropertyReferences=prop_reference_list,
            )

            read_access_spec_list.append(read_access_spec)

        if not read_access_spec_list:
            raise RuntimeError("at least one read access specification required")

        # build the request
        request = ReadPropertyMultipleRequest(
            listOfReadAccessSpecs=read_access_spec_list
        )
        request.pduDestination = Address(addr)
        return request

    def build_rpm_request_from_dict(self, request_dict, vendor_id):
        """
        Read property multiple allow to read a lot of properties with only one request
        The existing RPM function is made using a string that must be created using bacpypes 
        console style and is hard to automate.

        This new version will be an attempt to improve that::

            _rpm = {'address': '11:2',
                'objects': {'analogInput:1': ['presentValue', 'description', 'unit', 'objectList@idx:0'],
                            'analogInput:2': ['presentValue', 'description', 'unit', 'objectList@idx:0'],
                },
                vendor_id: 842
                }

        """
        vendor_id = 842
        addr = request_dict["address"]
        objects = request_dict["objects"]
        if "vendor_id" in request_dict.keys():
            vendor_id = int(request_dict["vendor_id"])

        read_access_spec_list = []

        for obj, list_of_properties in objects.items():
            obj_type, obj_instance = obj.split(":")
            obj_type = validate_object_type(obj_type, vendor_id=vendor_id)
            obj_instance = int(obj_instance)
            property_reference_list = build_property_reference_list(
                obj_type, list_of_properties
            )
            read_acces_spec = build_read_access_spec(
                obj_type, obj_instance, property_reference_list
            )
            read_access_spec_list.append(read_acces_spec)

        if not read_access_spec_list:
            raise RuntimeError("at least one read access specification required")

        # build the request
        request = ReadPropertyMultipleRequest(
            listOfReadAccessSpecs=read_access_spec_list
        )
        request.pduDestination = Address(addr)

        return request

    def build_rrange_request(
        self, args, range_params=None, arr_index=None, vendor_id=0, bacoid=None
    ):
        addr, obj_type, obj_inst, prop_id = args[:4]

        vendor_id = vendor_id
        bacoid = bacoid

        if obj_type.isdigit():
            obj_type = int(obj_type)
        elif not get_object_class(obj_type, vendor_id=vendor_id):
            raise ValueError("Unknown object type {}".format(obj_type))

        obj_inst = int(obj_inst)

        if prop_id.isdigit():
            prop_id = int(prop_id)
        datatype = get_datatype(obj_type, prop_id, vendor_id=vendor_id)
        if not datatype:
            raise ValueError("invalid property for object type")

        # build a request
        request = ReadRangeRequest(
            objectIdentifier=(obj_type, obj_inst), propertyIdentifier=prop_id
        )
        request.pduDestination = Address(addr)
        if range_params is not None:
            range_type, first, date, time, count = range_params
            if range_type == "p":
                rbp = RangeByPosition(referenceIndex=int(first), count=int(count))
                request.range = Range(byPosition=rbp)
            elif range_type == "s":
                rbs = RangeBySequenceNumber(
                    referenceSequenceNumber=int(first), count=int(count)
                )
                request.range = Range(bySequenceNumber=rbs)
            elif range_type == "t":
                rbt = RangeByTime(
                    referenceTime=DateTime(
                        date=Date(date).value, time=Time(time).value
                    ),
                    count=int(count),
                )
                request.range = Range(byTime=rbt)
            elif range_type == "x":
                # should be missing required parameter
                request.range = Range()
            else:
                raise ValueError("unknown range type: %r" % (range_type,))

        if len(args) == 5:
            request.propertyArrayIndex = int(args[4])
        self._log.debug("{:<20} {!r}".format("REQUEST", request))
        return request

    def readRange(
        self,
        args,
        range_params=None,
        arr_index=None,
        vendor_id=0,
        bacoid=None,
        timeout=10,
    ):
        """
        Build a ReadProperty request, wait for the answer and return the value

        :param args: String with <addr> <type> <inst> <prop> [ <indx> ]
        :returns: data read from device (str representing data like 10 or True)

        *Example*::

            import BAC0
            myIPAddr = '192.168.1.10/24'
            bacnet = BAC0.connect(ip = myIPAddr)
            bacnet.read('2:5 analogInput 1 presentValue')

        Requests the controller at (Network 2, address 5) for the presentValue of
        its analog input 1 (AI:1).
        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        args_split = args.split()

        self.log_title("Read range ", args_split)

        vendor_id = vendor_id
        bacoid = bacoid

        try:
            # build ReadProperty request
            request = self.build_rrange_request(
                args_split,
                range_params=range_params,
                arr_index=arr_index,
                vendor_id=vendor_id,
                bacoid=bacoid,
            )
            iocb = IOCB(request)
            iocb.set_timeout(timeout)
            # pass to the BACnet stack
            deferred(self.this_application.request_io, iocb)
            self._log.debug("{:<20} {!r}".format("iocb", iocb))

        except ReadRangeException as error:
            # construction error
            self._log.exception("exception: {!r}".format(error))

        iocb.wait()  # Wait for BACnet response

        if iocb.ioResponse:  # successful response
            apdu = iocb.ioResponse

            if not isinstance(apdu, ReadRangeACK):  # expecting an ACK
                self._log.warning("Not an ack, see debug for more infos.")
                self._log.debug(
                    "Not an ack. | APDU : {} / {}".format((apdu, type(apdu)))
                )
                return

            # find the datatype
            datatype = get_datatype(
                apdu.objectIdentifier[0], apdu.propertyIdentifier, vendor_id=vendor_id
            )
            if not datatype:
                # raise TypeError("unknown datatype")
                datatype = cast_datatype_from_tag(
                    apdu.propertyValue,
                    apdu.objectIdentifier[0],
                    apdu.propertyIdentifier,
                )

            try:
                value = apdu.itemData.cast_out(datatype)
            except TypeError as error:
                self._log.error(
                    "Problem casting value : {} | Datatype : {} | error : {}".format(
                        apdu.itemData, datatype, error
                    )
                )
                return apdu

            self._log.debug("{:<20} {:<20}".format("value", "datatype"))
            self._log.debug("{!r:<20} {!r:<20}".format(value, datatype))
            return value

        if iocb.ioError:  # unsuccessful: error/reject/abort
            apdu = iocb.ioError
            reason = find_reason(apdu)
            if reason == "segmentationNotSupported":
                self._log.warning(
                    "Segmentation not supported... will read properties one by one..."
                )
                self._log.debug("The Request was : {}".format(args_split))
                value = self._split_the_read_request(args, arr_index)
                return value
            else:
                if reason == "unknownProperty":
                    if "priorityArray" in args:
                        self._log.debug("Unknown property {}".format(args))
                    else:
                        self._log.warning("Unknown property {}".format(args))
                    if "description" in args:
                        return ""
                    elif "inactiveText" in args:
                        return "Off"
                    elif "activeText" in args:
                        return "On"
                    else:
                        raise UnknownPropertyError("Unknown property {}".format(args))
                elif reason == "unknownObject":
                    self._log.warning("Unknown object {}".format(args))
                    raise UnknownObjectError("Unknown object {}".format(args))
                else:
                    # Other error... consider NoResponseFromController (65)
                    # even if the realy reason is another one
                    raise NoResponseFromController(
                        "APDU Abort Reason : {}".format(reason)
                    )

    def read_priority_array(self, addr, obj, obj_instance):
        pa = self.read("{} {} {} priorityArray".format(addr, obj, obj_instance))
        res = [pa]
        for each in range(1, 17):
            _pa = pa[each]
            for k, v in _pa.__dict__.items():
                if v != None:
                    res.append(v)
        return res


def find_reason(apdu):
    try:
        if apdu == TimeoutError:
            return "Timeout"
        elif apdu.pduType == RejectPDU.pduType:
            reasons = RejectReason.enumerations
        elif apdu.pduType == AbortPDU.pduType:
            reasons = AbortReason.enumerations
        else:
            if apdu.errorCode and apdu.errorClass:
                return "{}".format(apdu.errorCode)
            else:
                raise ValueError("Cannot find reason...")
        code = apdu.apduAbortRejectReason
        try:
            return [k for k, v in reasons.items() if v == code][0]
        except IndexError:
            return code
    except KeyError as err:
        return "KeyError: {} has no key {0!r}".format(type(apdu), err.args[0])


def cast_datatype_from_tag(propertyValue, obj_id, prop_id):
    try:
        tag_list = propertyValue.tagList.tagList
        if tag_list[0].tagClass == 0:
            tag = tag_list[0].tagNumber

            datatype = Tag._app_tag_class[tag]
        else:
            from bacpypes.constructeddata import ArrayOf

            subtype_tag = propertyValue.tagList.tagList[0].tagList[0].tagNumber
            datatype = ArrayOf(Tag._app_tag_class[subtype_tag])
        value = {"{}_{}".format(obj_id, prop_id): propertyValue.cast_out(datatype)}
    except:
        value = {"{}_{}".format(obj_id, prop_id): propertyValue}
    return value


def validate_object_type(obj_type, vendor_id=842):
    if obj_type.isdigit():
        obj_type = int(obj_type)
    elif "@obj_" in obj_type:
        obj_type = int(obj_type.split("_")[1])
    elif not get_object_class(obj_type, vendor_id=vendor_id):
        raise ValueError("Unknown object type : {}".format(obj_type))
    return obj_type


def build_read_access_spec(obj_type, obj_instance, property_reference_list):
    return ReadAccessSpecification(
        objectIdentifier=(obj_type, obj_instance),
        listOfPropertyReferences=property_reference_list,
    )


def build_property_reference_list(obj_type, list_of_properties):
    property_reference_list = []
    for prop in list_of_properties:
        idx = None
        if "@idx:" in prop:
            prop, idx = prop.split("@idx:")
        prop_id = validate_property_id(obj_type, prop)
        prop_reference = PropertyReference(propertyIdentifier=prop_id)
        if idx:
            prop_reference.propertyArrayIndex = int(idx)
        property_reference_list.append(prop_reference)
    return property_reference_list


def validate_property_id(obj_type, prop_id):
    if prop_id in PropertyIdentifier.enumerations:
        if prop_id in (
            "all",
            "required",
            "optional",
            "objectName",
            "objectType",
            "objectIdentifier",
            "polarity",
        ):
            return prop_id
        elif validate_datatype(obj_type, prop_id) is not None:
            return prop_id
        else:
            raise ValueError(
                "invalid property for object type : {} | {}".format(obj_type, prop_id)
            )
    elif "@prop_" in prop_id:
        return int(prop_id.split("_")[1])
    # elif "@obj_" in prop_id:
    else:
        raise ValueError("{} is an invalid property for {}".format(prop_id, obj_type))


def validate_datatype(obj_type, prop_id, vendor_id=842):
    return get_datatype(obj_type, prop_id, vendor_id=vendor_id) if not None else False



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Complete Script - extended version of Lite Script

As everything is handled by the BasicScript, select the additional features you want::

    # Create a class that implements a basic script with read and write functions
    from BAC0.scripts.BasicScript import BasicScript
    from BAC0.core.io.Read import ReadProperty
    from BAC0.core.io.Write import WriteProperty
    class ReadWriteScript(BasicScript,ReadProperty,WriteProperty)

Once the class is created, create the local object and use it::

    bacnet = ReadWriteScript(localIPAddr = '192.168.1.10')
    bacnet.read('2:5 analogInput 1 presentValue)

"""
# --- standard Python modules ---
from datetime import datetime
import logging
import pandas as pd
import time


# --- 3rd party modules ---
from bokeh.application import Application

# --- this application's modules ---
from ..scripts.Lite import Lite

from ..core.io.IOExceptions import (
    BokehServerCantStart,
    NoResponseFromController,
    UnrecognizedService,
)
from ..core.utils.notes import note_and_log, update_log_level

from ..web.BokehRenderer import (
    DevicesTableHandler,
    DynamicPlotHandler,
    NotesTableHandler,
)
from ..web.BokehServer import Bokeh_Worker
from ..web.FlaskServer import FlaskServer

# ------------------------------------------------------------------------------


class Stats_Mixin:
    """
    Statistics used by Flask App
    """

    @property
    def number_of_devices(self):
        if not self.discoveredDevices:
            return 0
        s = []
        [s.append(x) for x in self.discoveredDevices.items() if x[1] > 0]
        return len(s)

    @property
    def number_of_registered_trends(self):
        if self.trends:
            return len(self.trends)
        else:
            return 0

    def number_of_devices_per_network(self):
        total = float(self.number_of_devices)
        if total == 0:
            return (["No Devices"], ["0"], ["0%%"])
        labels = ["IP"]
        series_pct = ["%.2f %%" % (len(self.network_stats["ip_devices"]) / total * 100)]
        series = [len(self.network_stats["ip_devices"]) / total * 100]
        for each in self.network_stats["mstp_map"].keys():
            labels.append("MSTP #{}".format(each))
            series_pct.append(
                "%.2f %%" % (len(self.network_stats["mstp_map"][each]) / total * 100)
            )
            series.append(len(self.network_stats["mstp_map"][each]) / total * 100)
        return (labels, series, series_pct)

    def print_list(self, lst):
        s = ""
        try:
            s += lst[0]
        except IndexError:
            return s
        try:
            for each in lst[1:]:
                s = s + ", " + each
        except IndexError:
            pass
        return s

    @property
    def network_stats(self):
        """
        Used by Flask to show informations on the network
        """
        statistics = {}
        mstp_networks = []
        mstp_map = {}
        ip_devices = []
        bacoids = []
        mstp_devices = []
        if self.discoveredDevices:
            for address, bacoid in self.discoveredDevices.keys():
                if ":" in address:
                    net, mac = address.split(":")
                    mstp_networks.append(net)
                    mstp_devices.append(mac)
                    try:
                        mstp_map[net].append(mac)
                    except KeyError:
                        mstp_map[net] = []
                        mstp_map[net].append(mac)
                else:
                    net = "ip"
                    mac = address
                    ip_devices.append(address)
                bacoids.append((bacoid, address))
        mstpnetworks = sorted(set(mstp_networks))
        statistics["mstp_networks"] = mstpnetworks
        statistics["ip_devices"] = sorted(ip_devices)
        statistics["bacoids"] = sorted(bacoids)
        statistics["mstp_map"] = mstp_map
        statistics["timestamp"] = str(datetime.now())
        statistics["number_of_devices"] = self.number_of_devices
        statistics["number_of_registered_devices"] = len(self.registered_devices)
        statistics["print_mstpnetworks"] = self.print_list(mstpnetworks)
        return statistics


@note_and_log
class Complete(Lite, Stats_Mixin):
    """
    Build a BACnet application to accept read and write requests.
    [Basic Whois/IAm functions are implemented in parent BasicScript class.]
    Once created, execute a whois() to build a list of available controllers.
    Initialization requires information on the local device.

    :param ip='127.0.0.1': Address must be in the same subnet as the BACnet network
        [BBMD and Foreign Device - not supported]

    :param bokeh_server: (boolean) If set to false, will prevent Bokeh server
        from being started. Can help troubleshoot issues with Bokeh. By default,
        set to True.
    """

    def __init__(
        self,
        ip=None,
        mask=None,
        port=None,
        bbmdAddress=None,
        bbmdTTL=0,
        ping=True,
        bokeh_server=True,
        flask_port=8111,
        **params
    ):
        Lite.__init__(
            self,
            ip=ip,
            mask=mask,
            port=port,
            bbmdAddress=bbmdAddress,
            bbmdTTL=bbmdTTL,
            ping=ping,
            **params
        )
        self.flask_port = flask_port
        if bokeh_server:
            self.start_bokeh()
            self.FlaskServer.start()
        else:
            self._log.warning("Bokeh server not started. Trend feature will not work")
        self.discover()

    @property
    def devices(self):
        lst = []
        for device in list(self.discoveredDevices):
            try:
                deviceName, vendorName = self.readMultiple(
                    "{} device {} objectName vendorName".format(device[0], device[1])
                )
            except UnrecognizedService:
                deviceName = self.read(
                    "{} device {} objectName".format(device[0], device[1])
                )
                vendorName = self.read(
                    "{} device {} vendorName".format(device[0], device[1])
                )
            except NoResponseFromController:
                self._log.info("No response from {}".format(device))
                continue
            lst.append((deviceName, vendorName, device[0], device[1]))
        df = pd.DataFrame(
            lst, columns=["Name", "Manufacturer", "Address", " Device ID"]
        ).set_index("Name")
        try:
            return df.sort_values("Address")
        except AttributeError:
            return df

    def start_bokeh(self):
        try:
            self.note("Starting Bokeh Serve")
            # Need to create the device document here
            devHandler = DevicesTableHandler(self)
            dev_app = Application(devHandler)
            trendHandler = DynamicPlotHandler(self)
            notesHandler = NotesTableHandler(self)
            self.trend_app = Application(trendHandler)
            self.notes_app = Application(notesHandler)
            self.bk_worker = Bokeh_Worker(
                dev_app, self.trend_app, self.notes_app, self.localIPAddr.addrTuple[0]
            )
            self.FlaskServer = FlaskServer(
                network=self, port=self.flask_port, ip=self.localIPAddr.addrTuple[0]
            )
            self.bk_worker.start()
            self.bokehserver = True
            time.sleep(1)
            try:
                logging.getLogger().handlers[0].setLevel(logging.CRITICAL)
            except IndexError:
                # Possibly in Jupyter Notebook... root logger not defined
                pass
            update_log_level("default", log_this=False)
            self._log.info(
                "Server started : http://{}:{}".format(
                    self.localIPAddr.addrTuple[0], self.flask_port
                )
            )

        except OSError as error:
            self.bokehserver = False
            self._log.error(
                "[bokeh serve] required for trending (controller.chart) features"
            )
            self._log.error(error)

        except RuntimeError as rterror:
            self.bokehserver = False
            self._log.warning("Server already running")

        except BokehServerCantStart:
            self.bokehserver = False
            self._log.error("No Bokeh Server - controller.chart not available")

    def __repr__(self):
        return "Bacnet Network using ip {} with device id {} | Featuring Bokeh and Pandas".format(
            self.localIPAddr, self.Boid
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
sql.py - 
"""

# --- standard Python modules ---
import pickle
import os.path

# --- 3rd party modules ---
import sqlite3
import contextlib

try:
    import pandas as pd
    from pandas.io import sql

    try:
        from pandas import Timestamp
    except ImportError:
        from pandas.lib import Timestamp
    _PANDAS = True
except ImportError:
    _PANDAS = False

from ..core.io.IOExceptions import RemovedPointException

# --- this application's modules ---

# ------------------------------------------------------------------------------


class SQLMixin(object):
    """
    Use SQL to persist a device's contents.  By saving the device contents to an SQL 
    database, you can work with the device's data while offline, or while the device 
    is not available. 
    """

    def _read_from_sql(self, request, db_name):
        """
        Using the contextlib, I hope to close the connection to database when
        not in use
        """
        with contextlib.closing(sqlite3.connect("{}.db".format(db_name))) as con:
            return sql.read_sql(sql=request, con=con)

    def dev_properties_df(self):
        dic = self.properties.asdict.copy()
        dic.pop("network", None)
        dic.pop("pss", None)
        return dic

    def points_properties_df(self):
        """
        Return a dictionary of point/point_properties in preparation for storage in SQL.
        """
        pprops = {}
        for each in self.points:
            p = each.properties.asdict.copy()
            p.pop("device", None)
            p.pop("network", None)
            p.pop("simulated", None)
            p.pop("overridden", None)
            pprops[each.properties.name] = p

        return pd.DataFrame(pprops)

    def backup_histories_df(self, resampling="1s"):
        """
        Build a dataframe of the point histories
        By default, dataframe will be resampled for 1sec intervals,
        NaN will be forward filled then backward filled. This way, no
        NaN values will remains and analytics will be easier.

        Please note that this can be disabled using resampling=False

        In the process of building the dataframe, analog values are
        resampled using the mean() function. So we have intermediate 
        results between to records.

        For binary values, we'll use .last() so we won't get a 0.5 value
        which means nothing in this context. 

        If saving a DB that already exists, previous resampling will survive
        the merge of old data and new data.
        """
        backup = {}
        if isinstance(resampling, str):
            resampling_needed = True
            resampling_freq = resampling
        elif resampling in [0, False]:
            resampling_needed = False

        # print(resampling, resampling_freq, resampling_needed)
        for point in self.points:
            try:
                if resampling_needed and "binary" in point.properties.type:
                    backup[point.properties.name] = (
                        point.history.replace(["inactive", "active"], [0, 1])
                        .resample(resampling_freq)
                        .last()
                    )
                elif resampling_needed and "analog" in point.properties.type:
                    backup[point.properties.name] = point.history.resample(
                        resampling_freq
                    ).mean()
                else:
                    backup[point.properties.name] = point.history.resample(
                        resampling_freq
                    ).last()

            except Exception as error:
                self._log.error(
                    "Error in resampling {} | {} (probably not enough points)".format(
                        point, error
                    )
                )
                if "binary" in point.properties.type:
                    backup[point.properties.name] = point.history.replace(
                        ["inactive", "active"], [0, 1]
                    )
                elif "analog" in point.properties.type:
                    backup[point.properties.name] = point.history.resample(
                        resampling_freq
                    ).mean()
                else:
                    backup[point.properties.name] = point.history

        df = pd.DataFrame(dict([(k, pd.Series(v)) for k, v in backup.items()]))
        if resampling_needed:
            return (
                df.resample(resampling_freq)
                .last()
                .fillna(method="ffill")
                .fillna(method="bfill")
            )
        else:
            return df

    def save(self, filename=None, resampling=None):
        """
        Save the point histories to sqlite3 database.  
        Save the device object properties to a pickle file so the device can be reloaded.

        Resampling : valid Pandas resampling frequency. If 0 or False, dataframe will not be resampled on save.
        """
        if filename:
            if ".db" in filename:
                filename = filename.split(".")[0]
            self.properties.db_name = filename
        else:
            self.properties.db_name = "{}".format(self.properties.name)

        if resampling is None:
            resampling = self.properties.save_resampling

        # Does file exist? If so, append data
        if os.path.isfile("{}.db".format(self.properties.db_name)):
            his = self._read_from_sql(
                'select * from "{}"'.format("history"), self.properties.db_name
            )
            his.index = his["index"].apply(Timestamp)
            try:
                last = his.index[-1]
                df_to_backup = self.backup_histories_df(resampling=resampling)[last:]
            except IndexError:
                df_to_backup = self.backup_histories_df(resampling=resampling)

        else:
            self._log.debug("Creating a new backup database")
            df_to_backup = self.backup_histories_df(resampling=resampling)

        # DataFrames that will be saved to SQL
        with contextlib.closing(
            sqlite3.connect("{}.db".format(self.properties.db_name))
        ) as con:
            try:
                data = pd.read_sql("SELECT * FROM history", con)
                df = pd.concat([data, df_to_backup], sort=True)
            except:
                df = df_to_backup

            sql.to_sql(
                df_to_backup,
                name="history",
                con=con,
                index_label="index",
                index=True,
                if_exists="append",
            )

        # Saving other properties to a pickle file...
        prop_backup = {"device": self.dev_properties_df()}
        prop_backup["points"] = self.points_properties_df()
        with open("{}.bin".format(self.properties.db_name), "wb") as file:
            pickle.dump(prop_backup, file)

        if self.properties.clear_history_on_save:
            self.clear_histories()

        self._log.info("Device saved to {}.db".format(self.properties.db_name))

    def points_from_sql(self, db_name):
        """
        Retrieve point list from SQL database
        """
        points = self._read_from_sql("SELECT * FROM history;", db_name)
        return list(points.columns.values)[1:]

    def his_from_sql(self, db_name, point):
        """
        Retrive point histories from SQL database
        """
        his = self._read_from_sql('select * from "{}"'.format("history", db_name))
        his.index = his["index"].apply(Timestamp)
        return his.set_index("index")[point]

    def value_from_sql(self, db_name, point):
        """
        Take last known value as the value
        """
        return self.his_from_sql(db_name, point).last_valid_index()

    def read_point_prop(self, device_name, point):
        """
        Points properties retrieved from pickle
        """
        with open("{}.bin".format(device_name), "rb") as file:
            try:
                _point = pickle.load(file)["points"][point]
            except KeyError:
                raise RemovedPointException(
                    "{} not found (probably deleted)".format(point)
                )
            return _point

    def read_dev_prop(self, device_name):
        """
        Device properties retrieved from pickle
        """
        self._log.debug("Reading prop from DB file")
        with open("{}.bin".format(device_name), "rb") as file:
            return pickle.load(file)["device"]



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Points.py - Definition of points so operations on Read results are more convenient.
"""

# --- standard Python modules ---
from datetime import datetime
from collections import namedtuple
import time

from bacpypes.primitivedata import (
    CharacterString,
    Null,
    Atomic,
    Integer,
    Unsigned,
    Real,
    Enumerated,
)

# --- 3rd party modules ---
try:
    import pandas as pd
    from pandas.io import sql

    try:
        from pandas import Timestamp
    except ImportError:
        from pandas.lib import Timestamp
    _PANDAS = True
except ImportError:
    _PANDAS = False

# --- this application's modules ---
from ...tasks.Poll import SimplePoll as Poll
from ...tasks.Match import Match, Match_Value
from ..io.IOExceptions import (
    NoResponseFromController,
    UnknownPropertyError,
    RemovedPointException,
    WritePropertyException,
)
from ..utils.notes import note_and_log


# ------------------------------------------------------------------------------


class PointProperties(object):
    """
    A container for point properties.
    """

    def __init__(self):
        self.device = None
        self.name = None
        self.type = None
        self.address = None
        self.description = None
        self.units_state = None
        self.simulated = (False, None)
        self.overridden = (False, None)
        self.priority_array = None
        self.history_size = None
        self.bacnet_properties = {}

    def __repr__(self):
        return "{}".format(self.asdict)

    @property
    def asdict(self):
        return self.__dict__


# ------------------------------------------------------------------------------


@note_and_log
class Point:
    """
    Represents a device BACnet point.  Used to NumericPoint, BooleanPoint and EnumPoints.

    Each point implements a history feature. Each time the point is read, its value (with timestamp)
    is added to a history table. Histories capture the changes to point values over time.
    """

    def __init__(
        self,
        device=None,
        pointType=None,
        pointAddress=None,
        pointName=None,
        description=None,
        presentValue=None,
        units_state=None,
        history_size=None,
    ):

        self._history = namedtuple("_history", ["timestamp", "value"])
        self.properties = PointProperties()

        self._polling_task = namedtuple("_polling_task", ["task", "running"])
        self._polling_task.task = None
        self._polling_task.running = False

        self._match_task = namedtuple("_match_task", ["task", "running"])
        self._match_task.task = None
        self._match_task.running = False

        self._history.timestamp = []
        self._history.value = [presentValue]
        self._history.timestamp.append(datetime.now())

        self.properties.history_size = history_size

        self.properties.device = device
        self.properties.name = pointName
        self.properties.type = pointType
        self.properties.address = pointAddress

        self.properties.description = description
        self.properties.units_state = units_state
        self.properties.simulated = (False, 0)
        self.properties.overridden = (False, 0)

        self.cov_registered = False

    @property
    def value(self):
        """
        Retrieve value of the point
        """
        try:
            res = self.properties.device.properties.network.read(
                "{} {} {} presentValue".format(
                    self.properties.device.properties.address,
                    self.properties.type,
                    str(self.properties.address),
                ),
                vendor_id=self.properties.device.properties.vendor_id,
            )
            # self._trend(res)
        except Exception as e:
            raise

        return res

    def read_priority_array(self):
        """
        Retrieve priority array of the point
        """
        if self.properties.priority_array != False:
            try:
                res = self.properties.device.properties.network.read(
                    "{} {} {} priorityArray".format(
                        self.properties.device.properties.address,
                        self.properties.type,
                        str(self.properties.address),
                    ),
                    vendor_id=self.properties.device.properties.vendor_id,
                )
                self.properties.priority_array = res
            except (ValueError, UnknownPropertyError):
                self.properties.priority_array = False
            except Exception as e:
                raise Exception(
                    "Problem reading : {} | {}".format(self.properties.name, e)
                )

    def read_property(self, prop):
        try:
            return self.properties.device.properties.network.read(
                "{} {} {} {}".format(
                    self.properties.device.properties.address,
                    self.properties.type,
                    str(self.properties.address),
                    prop,
                ),
                vendor_id=0,
            )
        except Exception as e:
            raise Exception("Problem reading : {} | {}".format(self.properties.name, e))

    def update_bacnet_properties(self):
        """
        Retrieve bacnet properties for this point
        To retrieve something general, forcing vendor id 0
        """
        try:
            res = self.properties.device.properties.network.readMultiple(
                "{} {} {} all".format(
                    self.properties.device.properties.address,
                    self.properties.type,
                    str(self.properties.address),
                ),
                vendor_id=0,
                prop_id_required=True,
            )
            for each in res:
                if not each:
                    continue
                v, prop = each
                self.properties.bacnet_properties[prop] = v

        except Exception as e:
            raise Exception("Problem reading : {} | {}".format(self.properties.name, e))

    @property
    def bacnet_properties(self):
        if not self.properties.bacnet_properties:
            self.update_bacnet_properties()
        return self.properties.bacnet_properties

    @property
    def is_overridden(self):
        self.read_priority_array()
        if self.properties.priority_array == False:
            return False
        if self.priority(8) or self.priority(1):
            self.properties.overridden = (True, self.value)
            return True
        else:
            return False

    def priority(self, priority=None):
        if self.properties.priority_array == False:
            return None

        self.read_priority_array()
        if not priority:
            return self.properties.priority_array.debug_contents()
        if priority < 1 or priority > 16:
            raise IndexError("Please provide priority to read (1-16)")

        else:
            pa = self.properties.priority_array.value[priority]
            try:
                key, value = zip(*pa.dict_contents().items())
                if key[0] == "null":
                    return None
                else:
                    return (key[0], value[0])
            except ValueError:
                return None

    def _trend(self, res):
        self._history.timestamp.append(datetime.now())
        self._history.value.append(res)
        if self.properties.history_size is None:
            return
        else:
            if self.properties.history_size < 1:
                self.properties.history_size = 1
            if len(self._history.timestamp) >= self.properties.history_size:
                try:
                    self._history.timestamp = self._history.timestamp[
                        -self.properties.history_size :
                    ]
                    self._history.value = self._history.value[
                        -self.properties.history_size :
                    ]
                    assert len(self._history.timestamp) == len(self._history.value)
                except Exception as e:
                    self._log.exception("Can't append to history")

    @property
    def units(self):
        """
        Should return units
        """
        raise Exception("Must be overridden")

    @property
    def lastValue(self):
        """
        returns: last value read
        """
        if _PANDAS:
            return self.history.dropna().iloc[-1]
        else:
            return self._history.value[-1]

    @property
    def history(self):
        """
        returns : (pd.Series) containing timestamp and value of all readings
        """
        if not _PANDAS:
            return dict(zip(self._history.timestamp, self._history.value))
        idx = self._history.timestamp.copy()
        his_table = pd.Series(index=idx, data=self._history.value[: len(idx)])
        del idx
        his_table.name = ("{}/{}").format(
            self.properties.device.properties.name, self.properties.name
        )
        his_table.units = self.properties.units_state
        if self.properties.name in self.properties.device.binary_states:
            his_table.states = "binary"
        elif self.properties.name in self.properties.device.multi_states:
            his_table.states = "multistates"
        else:
            his_table.states = "analog"
        his_table.description = self.properties.description

        his_table.datatype = self.properties.type
        return his_table

    def clear_history(self):
        self._history.timestamp = []
        self._history.value = []

    def chart(self, remove=False):
        """
        Add point to the bacnet trending list
        """
        if remove:
            self.properties.device.properties.network.remove_trend(self)
        else:
            self.properties.device.properties.network.add_trend(self)

    def __getitem__(self, key):
        """
        Way to get points... presentValue, status, flags, etc...

        :param key: state
        :returns: list of enum states
        """
        if str(key).lower() in ["unit", "units", "state", "states"]:
            key = "units_state"
        try:
            return getattr(self.properties, key)
        except AttributeError:
            try:
                if "@prop_" in key:
                    key = key.split("prop_")[1]
                return self.read_property(key)
            except Exception:
                raise ValueError("Cannot find property named {}".format(key))

    def write(self, value, *, prop="presentValue", priority=""):
        """
        Write to present value of a point

        :param value: (float) numeric value
        :param prop: (str) property to write. Default = presentValue
        :param priority: (int) priority to which write.

        """
        if priority != "":
            if (
                isinstance(float(priority), float)
                and float(priority) >= 1
                and float(priority) <= 16
            ):
                priority = "- {}".format(priority)
            else:
                raise ValueError("Priority must be a number between 1 and 16")

        try:
            self.properties.device.properties.network.write(
                "{} {} {} {} {} {}".format(
                    self.properties.device.properties.address,
                    self.properties.type,
                    self.properties.address,
                    prop,
                    value,
                    priority,
                ),
                vendor_id=self.properties.device.properties.vendor_id,
            )
        except NoResponseFromController:
            raise

        # Read after the write so history gets updated.
        self.value

    def default(self, value):
        self.write(value, prop="relinquishDefault")

    def sim(self, value, *, force=False):
        """
        Simulate a value.  Sets the Out_Of_Service property- to disconnect the point from the
        controller's control.  Then writes to the Present_Value.
        The point name is added to the list of simulated points (self.simPoints)

        :param value: (float) value to simulate
        """
        if (
            not self.properties.simulated[0]
            or self.properties.simulated[1] != value
            or force != False
        ):
            self.properties.device.properties.network.sim(
                "{} {} {} presentValue {}".format(
                    self.properties.device.properties.address,
                    self.properties.type,
                    str(self.properties.address),
                    str(value),
                )
            )
            self.properties.simulated = (True, value)

    def out_of_service(self):
        """
        Sets the Out_Of_Service property [to True].
        """
        self.properties.device.properties.network.out_of_service(
            "{} {} {}".format(
                self.properties.device.properties.address,
                self.properties.type,
                str(self.properties.address),
            )
        )
        self.properties.simulated = (True, None)

    def release(self):
        """
        Clears the Out_Of_Service property [to False] - so the controller regains control of the point.
        """
        self.properties.device.properties.network.release(
            "{} {} {}".format(
                self.properties.device.properties.address,
                self.properties.type,
                str(self.properties.address),
            )
        )
        self.properties.simulated = (False, None)

    def ovr(self, value):
        self.write(value, priority=8)
        self.properties.overridden = (True, value)

    def auto(self):
        self.write("null", priority=8)
        self.properties.overridden = (False, 0)

    def release_ovr(self):
        self.write("null", priority=1)
        self.write("null", priority=8)
        self.properties.overridden = (False, None)

    def _setitem(self, value):
        """
        Called by _set, will trigger right function depending on
        point type to write to the value and make tests.
        This is default behaviour of the point  :
        AnalogValue are written to
        AnalogOutput are overridden
        """
        if "characterstring" in self.properties.type:
            self.write(value)

        elif "Value" in self.properties.type:
            if str(value).lower() == "auto":
                raise ValueError(
                    "Value was not simulated or overridden, cannot release to auto"
                )
            # analog value must be written to
            self.write(value)

        elif "Output" in self.properties.type:
            # analog output must be overridden
            if str(value).lower() == "auto":
                self.auto()
            else:
                self.ovr(value)
        else:
            # input are left... must be simulated
            if str(value).lower() == "auto":
                self.release()
            else:
                self.sim(value)

    def _set(self, value):
        """
        Allows the syntax:
            device['point'] = value
        """
        raise NotImplementedError("Must be overridden")

    def poll(self, command="start", *, delay=10):
        """
        Poll a point every x seconds (delay=x sec)
        Stopped by using point.poll('stop') or .poll(0) or .poll(False)
        or by setting a delay = 0
        """
        if (
            str(command).lower() == "stop"
            or command == False
            or command == 0
            or delay == 0
        ):

            if isinstance(self._polling_task.task, Poll):
                self._polling_task.task.stop()
                self._polling_task.task = None
                self._polling_task.running = False

        elif self._polling_task.task is None:
            self._polling_task.task = Poll(self, delay=delay)
            self._polling_task.task.start()
            self._polling_task.running = True

        elif self._polling_task.running:
            self._polling_task.task.stop()
            self._polling_task.running = False
            self._polling_task.task = Poll(self, delay=delay)
            self._polling_task.task.start()
            self._polling_task.running = True

        else:
            raise RuntimeError("Stop polling before redefining it")

    def match(self, point, *, delay=5):
        """
        This allow functions like :
            device['status'].match('command')

        A fan status for example will follow the command...
        """
        if self._match_task.task is None:
            self._match_task.task = Match(command=point, status=self, delay=delay)
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay > 0:
            self._match_task.task.stop()
            self._match_task.running = False
            time.sleep(1)

            self._match_task.task = Match(command=point, status=self, delay=delay)
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay == 0:
            self._match_task.task.stop()
            self._match_task.running = False

        else:
            raise RuntimeError("Stop task before redefining it")

    def match_value(self, value, *, delay=5, use_last_value=False):
        """
        This allow functions like :
            device['point'].match('value')

        A sensor will follow a calculation...
        """
        if self._match_task.task is None:
            self._match_task.task = Match_Value(
                value=value, point=self, delay=delay, use_last_value=use_last_value
            )
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay > 0:
            self._match_task.task.stop()
            self._match_task.running = False
            time.sleep(1)

            self._match_task.task = Match_Value(value=value, point=self, delay=delay)
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay == 0:
            self._match_task.task.stop()
            self._match_task.running = False

        else:
            raise RuntimeError("Stop task before redefining it")

    def __len__(self):
        """
        Length of a point = # of history records
        """
        return len(self.history)

    def subscribe_cov(self, confirmed=True, lifetime=None, callback=None):
        address = self.properties.device.properties.address
        obj_tuple = (self.properties.type, int(self.properties.address))
        self.properties.device.properties.network.cov(
            address,
            obj_tuple,
            confirmed=confirmed,
            lifetime=lifetime,
            callback=callback,
        )

    def cancel_cov(self, callback=None):
        address = self.properties.device.properties.address
        obj_tuple = (self.properties.type, int(self.properties.address))
        self.properties.device.properties.network.cancel_cov(
            address, obj_tuple, callback=callback
        )


# ------------------------------------------------------------------------------


class NumericPoint(Point):
    """
    Representation of a Numeric value
    """

    def __init__(
        self,
        device=None,
        pointType=None,
        pointAddress=None,
        pointName=None,
        description=None,
        presentValue=None,
        units_state=None,
        history_size=None,
    ):

        Point.__init__(
            self,
            device=device,
            pointType=pointType,
            pointAddress=pointAddress,
            pointName=pointName,
            description=description,
            presentValue=presentValue,
            units_state=units_state,
            history_size=history_size,
        )

    @property
    def value(self):
        res = super().value
        self._trend(res)
        return res

    @property
    def units(self):
        return self.properties.units_state

    def _set(self, value):
        if str(value).lower() == "auto":
            self._setitem(value)
        else:
            try:
                if isinstance(value, Point):
                    value = value.lastValue
                val = float(value)
                if isinstance(val, float):
                    self._setitem(value)
            except Exception as error:
                raise WritePropertyException(
                    "Problem writing to device : {}".format(error)
                )

    def __repr__(self):
        try:
            polling = self.properties.device.properties.pollDelay
            if (polling < 90 and polling > 0) or self.cov_registered:
                val = float(self.lastValue)
            else:
                val = float(self.value)
        except ValueError:
            self._log.error(
                "Cannot convert value {}. Device probably disconnected".format(
                    self.value
                )
            )
            # Probably disconnected
            val = None
        return "{}/{} : {:.2f} {}".format(
            self.properties.device.properties.name,
            self.properties.name,
            val,
            self.properties.units_state,
        )

    def __add__(self, other):
        return self.value + other

    def __sub__(self, other):
        return self.value - other

    def __mul__(self, other):
        return self.value * other

    def __truediv__(self, other):
        return self.value / other

    def __lt__(self, other):
        return self.value < other

    def __le__(self, other):
        return self.value <= other

    def __eq__(self, other):
        return self.value == other

    def __gt__(self, other):
        return self.value > other

    def __ge__(self, other):
        return self.value >= other


# ------------------------------------------------------------------------------


class BooleanPoint(Point):
    """
    Representation of a Boolean value
    """

    def __init__(
        self,
        device=None,
        pointType=None,
        pointAddress=None,
        pointName=None,
        description=None,
        presentValue=None,
        units_state=None,
        history_size=None,
    ):

        Point.__init__(
            self,
            device=device,
            pointType=pointType,
            pointAddress=pointAddress,
            pointName=pointName,
            description=description,
            presentValue=presentValue,
            units_state=units_state,
            history_size=history_size,
        )

    def _trend(self, res):
        res = "1: active" if res == "active" else "0: inactive"
        super()._trend(res)

    @property
    def value(self):
        """
        Read the value from BACnet network
        """
        res = super().value
        self._trend(res)

        if res == "inactive":
            self._key = 0
            self._boolKey = False
        else:
            self._key = 1
            self._boolKey = True
        return res

    @property
    def boolValue(self):
        """
        returns : (boolean) Value
        """
        if ":" in self.lastValue:
            _val = int(self.lastValue.split(":")[0])
        else:
            _val = self.lastValue
        if _val in [1, "active"]:
            self._key = 1
            self._boolKey = True
        else:
            self._key = 0
            self._boolKey = False
        return self._boolKey

    @property
    def units(self):
        """
        Boolean points don't have units
        """
        return None

    def _set(self, value):
        try:
            if value == True:
                self._setitem("active")
            elif value == False:
                self._setitem("inactive")
            elif str(value) in ["inactive", "active"] or str(value).lower() == "auto":
                self._setitem(value)
            else:
                raise ValueError(
                    'Value must be boolean True, False or "active"/"inactive"'
                )
        except (Exception, ValueError) as error:
            raise WritePropertyException("Problem writing to device : {}".format(error))

    def __repr__(self):
        polling = self.properties.device.properties.pollDelay
        if (polling >= 90 or polling <= 0) and not self.cov_registered:
            # Force reading
            self.value
        return "{}/{} : {}".format(
            self.properties.device.properties.name, self.properties.name, self.boolValue
        )

    def __or__(self, other):
        return self.boolValue | other

    def __and__(self, other):
        return self.boolValue & other

    def __xor__(self, other):
        return self.boolValue ^ other

    def __eq__(self, other):
        return self.boolValue == other


# ------------------------------------------------------------------------------


class EnumPoint(Point):
    """
    Representation of an Enumerated (multiState) value
    """

    def __init__(
        self,
        device=None,
        pointType=None,
        pointAddress=None,
        pointName=None,
        description=None,
        presentValue=None,
        units_state=None,
        history_size=None,
    ):

        Point.__init__(
            self,
            device=device,
            pointType=pointType,
            pointAddress=pointAddress,
            pointName=pointName,
            description=description,
            presentValue=presentValue,
            units_state=units_state,
            history_size=history_size,
        )

    def _trend(self, res):
        res = "{}: {}".format(res, self.get_state(res))
        super()._trend(res)

    @property
    def value(self):
        res = super().value
        # self._log.info("Value : {}".format(res))
        # self._log.info("EnumValue : {}".format(self.get_state(res)))
        self._trend(res)
        return res

    def get_state(self, v):
        try:
            return self.properties.units_state[v - 1]
        except TypeError:
            return "n/a"

    @property
    def enumValue(self):
        """
        returns: (str) Enum state value
        """
        try:
            if ":" in self.lastValue:
                _val = int(self.lastValue.split(":")[0])
                return self.get_state(_val)
        except TypeError:
            # probably first occurence of history... retry
            return self.get_state(self.lastValue)
        except IndexError:
            value = "unknown"
        except ValueError:
            value = "NaN"
        return value

    @property
    def units(self):
        """
        Enums have 'state text' instead of units.
        """
        return None

    def _set(self, value):
        try:
            if isinstance(value, int):
                self._setitem(value)
            elif str(value) in self.properties.units_state:
                self._setitem(self.properties.units_state.index(value) + 1)
            elif str(value).lower() == "auto":
                self._setitem("auto")
            else:
                raise ValueError(
                    "Value must be integer or correct enum state : {}".format(
                        self.properties.units_state
                    )
                )
        except (Exception, ValueError) as error:
            raise WritePropertyException("Problem writing to device : {}".format(error))

    def __repr__(self):
        polling = self.properties.device.properties.pollDelay
        if (polling >= 90 or polling <= 0) and not self.cov_registered:
            # Force reading
            self.value
        return "{}/{} : {}".format(
            self.properties.device.properties.name, self.properties.name, self.enumValue
        )

    def __eq__(self, other):
        return self.value == self.properties.units_state.index(other) + 1


class StringPoint(Point):
    """
    Representation of CharacterString value
    """

    def __init__(
        self,
        device=None,
        pointType=None,
        pointAddress=None,
        pointName=None,
        description=None,
        units_state=None,
        presentValue=None,
        history_size=None,
    ):

        Point.__init__(
            self,
            device=device,
            pointType=pointType,
            pointAddress=pointAddress,
            pointName=pointName,
            description=description,
            presentValue=presentValue,
            history_size=history_size,
        )

    @property
    def units(self):
        """
        Characterstring value do not have units or state text
        """
        return None

    def _trend(self, res):
        super()._trend(res)

    @property
    def value(self):
        res = super().value
        self._trend(res)
        return res

    def _set(self, value):
        try:
            if isinstance(value, str):
                self._setitem(value)
            elif isinstance(value, CharacterString):
                self._setitem(value.value)
            else:
                raise ValueError("Value must be string or CharacterString")
        except (Exception, ValueError) as error:
            raise WritePropertyException("Problem writing to device : {}".format(error))

    def __repr__(self):
        try:
            polling = self.properties.device.properties.pollDelay
            if (polling < 90 and polling > 0) or self.cov_registered:
                val = str(self.lastValue)
            else:
                val = str(self.value)
        except ValueError:
            self._log.error("Cannot convert value. Device probably disconnected")
            # Probably disconnected
            val = None
        return "{}/{} : {}".format(
            self.properties.device.properties.name, self.properties.name, val
        )

    def __eq__(self, other):
        return self.value == other.value


# ------------------------------------------------------------------------------


class OfflinePoint(Point):
    """
    When offline (DB state), points needs to behave in a particular way
    (we can't read on bacnet...)
    """

    def __init__(self, device, name):
        self.properties = PointProperties()
        self.properties.device = device
        dev_name = self.properties.device.properties.db_name
        try:
            props = self.properties.device.read_point_prop(dev_name, name)
        except RemovedPointException:
            raise

        self.properties.name = props["name"]
        self.properties.type = props["type"]
        self.properties.address = props["address"]

        self.properties.description = props["description"]
        self.properties.units_state = props["units_state"]
        self.properties.simulated = "Offline"
        self.properties.overridden = "Offline"

        if "analog" in self.properties.type:
            self.new_state(NumericPointOffline)
        elif "multi" in self.properties.type:
            self.new_state(EnumPointOffline)
        elif "binary" in self.properties.type:
            self.new_state(BooleanPointOffline)
        elif "string" in self.properties.type:
            self.new_state(StringPointOffline)
        else:
            raise TypeError("Unknown point type")

    def new_state(self, newstate):
        self.__class__ = newstate


class NumericPointOffline(NumericPoint):
    @property
    def history(self):
        his = self.properties.device._read_from_sql(
            'select * from "{}"'.format("history"),
            self.properties.device.properties.db_name,
        )
        his.index = his["index"].apply(Timestamp)
        return his.set_index("index")[self.properties.name]

    @property
    def value(self):
        """
        Take last known value as the value
        """
        try:
            value = self.lastValue
        except IndexError:
            value = 65535
        return value

    def write(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def sim(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def release(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    @property
    def units(self):
        return self.properties.units_state

    def _set(self, value):
        raise OfflineException("Must be online to write")

    def __repr__(self):
        return "{}/{} : {:.2f} {}".format(
            self.properties.device.properties.name,
            self.properties.name,
            self.value,
            self.properties.units_state,
        )


class BooleanPointOffline(BooleanPoint):
    @property
    def history(self):
        his = self.properties.device._read_from_sql(
            'select * from "{}"'.format("history"),
            self.properties.device.properties.db_name,
        )
        his.index = his["index"].apply(Timestamp)
        return his.set_index("index")[self.properties.name]

    @property
    def value(self):
        try:
            value = self.lastValue
        except IndexError:
            value = "NaN"
        return value

    def _set(self, value):
        raise OfflineException("Point must be online to write")

    def write(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def sim(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def release(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")


class EnumPointOffline(EnumPoint):
    @property
    def history(self):
        his = self.properties.device._read_from_sql(
            'select * from "{}"'.format("history"),
            self.properties.device.properties.db_name,
        )
        his.index = his["index"].apply(Timestamp)
        return his.set_index("index")[self.properties.name]

    @property
    def value(self):
        """
        Take last known value as the value
        """
        try:
            value = self.lastValue
        except IndexError:
            value = "NaN"
        except ValueError:
            value = "NaN"
        return value

    @property
    def enumValue(self):
        """
        returns: (str) Enum state value
        """
        try:
            value = self.properties.units_state[int(self.lastValue) - 1]
        except IndexError:
            value = "unknown"
        except ValueError:
            value = "NaN"
        return value

    def _set(self, value):
        raise OfflineException("Point must be online to write")

    def write(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def sim(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def release(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")


class StringPointOffline(EnumPoint):
    @property
    def history(self):
        his = self.properties.device._read_from_sql(
            'select * from "{}"'.format("history"),
            self.properties.device.properties.db_name,
        )
        his.index = his["index"].apply(Timestamp)
        return his.set_index("index")[self.properties.name]

    @property
    def value(self):
        """
        Take last known value as the value
        """
        try:
            value = self.lastValue
        except IndexError:
            value = "NaN"
        except ValueError:
            value = "NaN"
        return value

    def _set(self, value):
        raise OfflineException("Point must be online to write")

    def write(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def sim(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")

    def release(self, value, *, prop="presentValue", priority=""):
        raise OfflineException("Must be online to write")


class OfflineException(Exception):
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Points.py - Definition of points so operations on Read results are more convenient.
"""

# --- standard Python modules ---
from datetime import datetime
from collections import namedtuple
import time
from itertools import islice


# --- 3rd party modules ---
try:
    import pandas as pd
    from pandas.io import sql

    try:
        from pandas import Timestamp
    except ImportError:
        from pandas.lib import Timestamp
    _PANDAS = True
except ImportError:
    _PANDAS = False

from bacpypes.object import TrendLogObject
from bacpypes.primitivedata import Date, Time

# --- this application's modules ---
from ...tasks.Poll import SimplePoll as Poll
from ...tasks.Match import Match, Match_Value
from ..io.IOExceptions import NoResponseFromController, UnknownPropertyError
from ..utils.notes import note_and_log


# ------------------------------------------------------------------------------


class TrendLogProperties(object):
    """
    A container for trend properties
    """

    def __init__(self):
        self.device = None
        self.oid = None
        self.object_name = None
        self.description = ""
        self.log_device_object_property = None
        self.buffer_size = 0
        self.record_count = 0
        self.total_record_count = 0
        self.log_interval = 0
        self.description = None
        self.statusFlags = None
        self.status_flags = {
            "in_alarm": False,
            "fault": False,
            "overridden": False,
            "out_of_service": False,
        }
        self._history_components = {"index": [], "logdatum": [], "status": []}
        self._df = None
        self.type = "TrendLog"
        self.units_state = "None"

    def __repr__(self):
        return "{} | Descr : {} | Record count : {}".format(
            self.object_name, self.description, self.record_count
        )

    @property
    def name(self):
        return self.object_name


@note_and_log
class TrendLog(TrendLogProperties):
    """
    BAC0 simplification of TrendLog Object
    """

    def __init__(
        self, OID, device=None, read_log_on_creation=True, multiple_request=None
    ):
        self.properties = TrendLogProperties()
        self.properties.device = device
        self.properties.oid = OID
        self.update_properties()

        if read_log_on_creation:
            self.read_log_buffer()
        self._last_index = 0

    def update_properties(self):
        try:
            self.properties.object_name, self.properties.description, self.properties.record_count, self.properties.buffer_size, self.properties.total_record_count, self.properties.log_device_object_property, self.properties.statusFlags, self.properties.log_interval = self.properties.device.properties.network.readMultiple(
                "{addr} trendLog {oid} objectName description recordCount bufferSize totalRecordCount logDeviceObjectProperty statusFlags logInterval".format(
                    addr=self.properties.device.properties.address,
                    oid=str(self.properties.oid),
                )
            )
        except Exception as error:
            raise Exception("Problem reading trendLog informations: {}".format(error))

    def _total_record_count(self):
        self.properties.total_record_count = self.properties.device.properties.network.read(
            "{addr} trendLog {oid} totalRecordCount".format(
                addr=self.properties.device.properties.address,
                oid=str(self.properties.oid),
            )
        )
        return self.properties.total_record_count

    def read_log_buffer(self):
        RECORDS = 10
        log_buffer = set()
        _actual_index = self._total_record_count()
        start = max(_actual_index - self.properties.record_count, self._last_index) + 1
        _count = _actual_index - start
        steps = int(_count / 10) + int(_count % 10)

        self._log.debug("Reading log : {} {} {}".format(start, _count, steps))

        _from = start
        for each in range(steps):
            range_params = ("s", _from, Date("1979-01-01"), Time("00:00"), RECORDS)
            _chunk = self.properties.device.properties.network.readRange(
                "{} trendLog {} logBuffer".format(
                    self.properties.device.properties.address, str(self.properties.oid)
                ),
                range_params=range_params,
            )
            _from += len(_chunk)
            for chunk in _chunk:
                log_buffer.add(chunk)

        self._last_index = _from
        self.create_dataframe(log_buffer)

    def create_dataframe(self, log_buffer):
        for each in log_buffer:
            year, month, day, dow = each.timestamp.date
            year = year + 1900
            hours, minutes, seconds, ms = each.timestamp.time
            self.properties._history_components["index"].append(
                pd.to_datetime(
                    "{}-{}-{} {}:{}:{}.{}".format(
                        year, month, day, hours, minutes, seconds, ms
                    ),
                    format="%Y-%m-%d %H:%M:%S.%f",
                )
            )
            self.properties._history_components["logdatum"].append(
                each.logDatum.dict_contents()
            )
            self.properties._history_components["status"].append(each.statusFlags)

        if _PANDAS:
            df = pd.DataFrame(
                {
                    "index": self.properties._history_components["index"],
                    "logdatum": self.properties._history_components["logdatum"],
                    "status": self.properties._history_components["status"],
                }
            )
            df = df.set_index("index")
            df["choice"] = df["logdatum"].apply(lambda x: list(x.keys())[0])
            df[self.properties.object_name] = df["logdatum"].apply(
                lambda x: list(x.values())[0]
            )

            self.properties._df = df
        else:
            # self.properties._history_components = (self.index, self.logdatum, self.status)
            self._log.warning(
                "Pandas not installed. Treating histories as simple list."
            )

    @property
    def history(self):
        self.read_log_buffer()
        if not _PANDAS:
            return dict(
                zip(
                    self.properties._history_components["index"],
                    self.properties._history_components["logdatum"],
                )
            )
        objectType, objectAddress = (
            self.properties.log_device_object_property.objectIdentifier
        )
        try:
            logged_point = self.properties.device.find_point(objectType, objectAddress)
        except ValueError:
            logged_point = None
        serie = self.properties._df[self.properties.object_name].copy()
        serie.units = logged_point.properties.units_state if logged_point else "n/a"
        serie.name = ("{}/{}").format(
            self.properties.device.properties.name, self.properties.object_name
        )
        if not logged_point:
            serie.states = "unknown"
        else:
            if logged_point.properties.name in self.properties.device.binary_states:
                serie.states = "binary"
            elif logged_point.properties.name in self.properties.device.multi_states:
                serie.states = "multistates"
            else:
                serie.states = "analog"
        serie.description = self.properties.description
        serie.datatype = objectType
        return serie.sort_index()

    def chart(self, remove=False):
        """
        Add point to the bacnet trending list
        """
        if not _PANDAS:
            self._log.error(
                "Pandas must be installed to use live chart feature. See documentation how how to run BAC0 in complete mode"
            )
        else:
            if remove:
                self.properties.device.properties.network.remove_trend(self)
            else:
                self.properties.device.properties.network.add_trend(self)

    def __repr__(self):
        return self.properties.__repr__()
